"""
This script generates the schema for the schema - e.g.
a definition of what inputs can be provided to `SchemaValidator()`.

The schema is generated from `pydantic_core/_types.py`.
"""
import importlib.util
import re
from collections.abc import Callable
from datetime import date, datetime, time, timedelta
from pathlib import Path
from typing import Any, Dict, ForwardRef, List, Type, Union

from black import Mode, TargetVersion, format_file_contents
from typing_extensions import get_args, is_typeddict

try:
    from typing import get_origin
except ImportError:

    def get_origin(t):
        return getattr(t, '__origin__', None)


THIS_DIR = Path(__file__).parent
SAVE_PATH = THIS_DIR / 'src' / 'self_schema.py'

# can't import _types.py directly as pydantic-core might not be installed
core_types_spec = importlib.util.spec_from_file_location('_typing', str(THIS_DIR / 'pydantic_core' / '_types.py'))
core_types = importlib.util.module_from_spec(core_types_spec)
core_types_spec.loader.exec_module(core_types)

# the validator for referencing schema (Schema is used recursively, so has to use a reference)
schema_ref_validator = {'type': 'recursive-ref', 'schema_ref': 'root-schema'}


def get_schema(obj):
    if isinstance(obj, str):
        return obj
    elif obj in (datetime, timedelta, date, time, bool, int, float, str):
        return obj.__name__
    elif is_typeddict(obj):
        return type_dict_schema(obj)
    elif obj == Any:
        return 'any'
    elif obj == type:
        # todo
        return 'any'

    origin = get_origin(obj)
    assert origin is not None, f'origin cannot be None, obj={obj}, you probably need to fix generate_self_schema.py'
    if origin is Union:
        return union_schema(obj)
    elif obj is Callable or origin is Callable:
        return 'callable'
    elif origin is core_types.Literal:
        expected = all_literal_values(obj)
        assert expected, f'literal "expected" cannot be empty, obj={obj}'
        return {'type': 'literal', 'expected': expected}
    elif issubclass(origin, List):
        return {'type': 'list', 'items_schema': get_schema(obj.__args__[0])}
    elif issubclass(origin, Dict):
        return {
            'type': 'dict',
            'keys_schema': get_schema(obj.__args__[0]),
            'values_schema': get_schema(obj.__args__[1]),
        }
    elif issubclass(origin, Type):
        # can't really use 'is-instance' since this is used for the class_ parameter of
        # 'is-instance' validators
        return 'any'
    else:
        # debug(obj)
        raise TypeError(f'Unknown type: {obj!r}')


def type_dict_schema(typed_dict):
    required_keys = getattr(typed_dict, '__required_keys__', set())
    fields = {}

    for field_name, field_type in typed_dict.__annotations__.items():
        required = field_name in required_keys
        schema = None
        if type(field_type) == ForwardRef:
            fr_arg = field_type.__forward_arg__
            fr_arg, matched = re.subn(r'NotRequired\[(.+)]', r'\1', fr_arg)
            if matched:
                required = False

            fr_arg, matched = re.subn(r'Required\[(.+)]', r'\1', fr_arg)
            if matched:
                required = True

            if 'Schema' == fr_arg or re.search('[^a-zA-Z]Schema', fr_arg):
                if fr_arg == 'Schema':
                    schema = schema_ref_validator
                elif fr_arg == 'List[Schema]':
                    schema = {'type': 'list', 'items_schema': schema_ref_validator}
                elif fr_arg == 'Dict[str, Schema]':
                    schema = {'type': 'dict', 'keys_schema': 'str', 'values_schema': schema_ref_validator}
                else:
                    raise ValueError(f'Unknown Schema forward ref: {fr_arg}')
            else:
                field_type = eval_forward_ref(field_type)

        if schema is None:
            if get_origin(field_type) == core_types.Required:
                required = True
                field_type = field_type.__args__[0]
            if get_origin(field_type) == core_types.NotRequired:
                required = False
                field_type = field_type.__args__[0]

            schema = get_schema(field_type)

        fields[field_name] = {'schema': schema, 'required': required}

    return {'type': 'typed-dict', 'description': typed_dict.__name__, 'fields': fields, 'extra_behavior': 'forbid'}


def union_schema(union_type):
    return {'type': 'union', 'choices': [get_schema(arg) for arg in union_type.__args__]}


def all_literal_values(type_):
    if get_origin(type_) is core_types.Literal:
        values = get_args(type_)
        return [x for value in values for x in all_literal_values(value)]
    else:
        return [type_]


def eval_forward_ref(type_):
    try:
        return type_._evaluate(core_types.__dict__, None, set())
    except TypeError:
        # for older python (3.7 at least)
        return type_._evaluate(core_types.__dict__, None)


def main():
    schema_union = core_types.Schema
    assert get_origin(schema_union) is Union, 'expected pydantic_core._types.Schema to be a union'

    schema = {
        'type': 'tagged-union',
        'ref': 'root-schema',
        'discriminator': 'self-schema-discriminator',
        'choices': {'plain-string': get_schema(schema_union.__args__[0])},
    }
    for s in schema_union.__args__[1:]:
        type_ = s.__annotations__['type']
        m = re.search(r"Literal\['(.+?)']", type_.__forward_arg__)
        assert m, f'Unknown schema type: {type_}'
        key = m.group(1)
        value = get_schema(s)
        if key == 'function' and value['fields']['mode']['schema']['expected'] == ['plain']:
            key = 'function-plain'
        elif key == 'tuple':
            if value['fields']['mode']['schema']['expected'] == ['positional']:
                key = 'tuple-positional'
            else:
                key = 'tuple-variable'

        schema['choices'][key] = value

    python_code = (
        f'# this file is auto-generated by generate_self_schema.py, DO NOT edit manually\nself_schema = {schema}\n'
    )
    mode = Mode(
        line_length=120,
        string_normalization=False,
        magic_trailing_comma=False,
        target_versions={TargetVersion.PY37, TargetVersion.PY38, TargetVersion.PY39, TargetVersion.PY310},
    )
    python_code = format_file_contents(python_code, fast=False, mode=mode)
    SAVE_PATH.write_text(python_code)
    print(f'Self schema definition written to {SAVE_PATH}')


if __name__ == '__main__':
    main()



"""
Numerous benchmarks of specific functionality.
"""
import json
import os
import platform
from datetime import date, datetime, timedelta, timezone
from typing import Dict, FrozenSet, List, Optional, Set, Union

import pytest

from pydantic_core import PydanticValueError, SchemaValidator, ValidationError, ValidationError as CoreValidationError

if os.getenv('BENCHMARK_VS_PYDANTIC'):
    try:
        from pydantic import BaseModel, StrictBool, StrictInt, StrictStr, ValidationError as PydanticValidationError
    except ImportError:
        BaseModel = None
else:
    BaseModel = None

skip_pydantic = pytest.mark.skipif(BaseModel is None, reason='skipping benchmarks vs. pydantic')


class TestBenchmarkSimpleModel:
    @pytest.fixture(scope='class')
    def pydantic_model(self):
        class PydanticModel(BaseModel):
            name: str
            age: int
            friends: List[int]
            settings: Dict[str, float]

        return PydanticModel

    @pytest.fixture(scope='class')
    def core_validator_fs(self):
        class CoreModel:
            __slots__ = '__dict__', '__fields_set__'

        return SchemaValidator(
            {
                'type': 'new-class',
                'class_type': CoreModel,
                'schema': {
                    'type': 'typed-dict',
                    'return_fields_set': True,
                    'fields': {
                        'name': {'schema': {'type': 'str'}},
                        'age': {'schema': {'type': 'int'}},
                        'friends': {'schema': {'type': 'list', 'items_schema': 'int'}},
                        'settings': {'schema': {'type': 'dict', 'keys_schema': 'str', 'values_schema': 'float'}},
                    },
                },
            }
        )

    @pytest.fixture(scope='class')
    def core_validator_not_fs(self):
        class CoreModel:
            __slots__ = '__dict__', '__fields_set__'

        return SchemaValidator(
            {
                'type': 'new-class',
                'class_type': CoreModel,
                'schema': {
                    'type': 'typed-dict',
                    'fields': {
                        'name': {'schema': {'type': 'str'}},
                        'age': {'schema': {'type': 'int'}},
                        'friends': {'schema': {'type': 'list', 'items_schema': 'int'}},
                        'settings': {'schema': {'type': 'dict', 'keys_schema': 'str', 'values_schema': 'float'}},
                    },
                },
            }
        )

    data = {'name': 'John', 'age': 42, 'friends': list(range(200)), 'settings': {f'v_{i}': i / 2.0 for i in range(50)}}

    @skip_pydantic
    @pytest.mark.benchmark(group='simple model - python')
    def test_pyd_python(self, pydantic_model, benchmark):
        benchmark(pydantic_model.parse_obj, self.data)

    @pytest.mark.benchmark(group='simple model - python')
    def test_core_python_fs(self, core_validator_fs, benchmark):
        m = core_validator_fs.validate_python(self.data)
        assert m.name == 'John'
        assert m.__dict__.keys() == {'name', 'age', 'friends', 'settings'}
        assert m.__fields_set__ == {'name', 'age', 'friends', 'settings'}
        benchmark(core_validator_fs.validate_python, self.data)

    @pytest.mark.benchmark(group='simple model - python')
    def test_core_python_not_fs(self, core_validator_not_fs, benchmark):
        m = core_validator_not_fs.validate_python(self.data)
        assert m.name == 'John'
        assert m.__dict__.keys() == {'name', 'age', 'friends', 'settings'}
        assert not hasattr(m, '__fields_set__')
        benchmark(core_validator_not_fs.validate_python, self.data)

    @skip_pydantic
    @pytest.mark.benchmark(group='simple model - JSON')
    def test_pyd_json(self, pydantic_model, benchmark):
        json_data = json.dumps(self.data)

        @benchmark
        def pydantic_json():
            obj = json.loads(json_data)
            return pydantic_model.parse_obj(obj)

    @pytest.mark.benchmark(group='simple model - JSON')
    def test_core_json_fs(self, core_validator_fs, benchmark):
        json_data = json.dumps(self.data)
        benchmark(core_validator_fs.validate_json, json_data)

    @pytest.mark.benchmark(group='simple model - JSON')
    def test_core_json_not_fs(self, core_validator_not_fs, benchmark):
        json_data = json.dumps(self.data)
        benchmark(core_validator_not_fs.validate_json, json_data)


bool_cases = [True, False, 0, 1, '0', '1', 'true', 'false', 'True', 'False']


@skip_pydantic
@pytest.mark.benchmark(group='bool')
def test_bool_pyd(benchmark):
    class PydanticModel(BaseModel):
        value: bool

    @benchmark
    def t():
        for case in bool_cases:
            PydanticModel(value=case)


@pytest.mark.benchmark(group='bool')
def test_bool_core(benchmark):
    schema_validator = SchemaValidator({'type': 'bool'})

    @benchmark
    def t():
        for case in bool_cases:
            schema_validator.validate_python(case)


small_class_data = {'name': 'John', 'age': 42}


@skip_pydantic
@pytest.mark.benchmark(group='create small model')
def test_small_class_pyd(benchmark):
    class PydanticModel(BaseModel):
        name: str
        age: int

    benchmark(PydanticModel.parse_obj, small_class_data)


@pytest.mark.benchmark(group='create small model')
def test_small_class_core_dict(benchmark):
    model_schema = {
        'type': 'typed-dict',
        'fields': {'name': {'schema': {'type': 'str'}}, 'age': {'schema': {'type': 'int'}}},
    }
    dict_schema_validator = SchemaValidator(model_schema)
    benchmark(dict_schema_validator.validate_python, small_class_data)


@pytest.mark.benchmark(group='create small model')
def test_small_class_core_model(benchmark):
    class MyCoreModel:
        # this is not required, but it avoids `__fields_set__` being included in `__dict__`
        __slots__ = '__dict__', '__fields_set__'
        # these are here just as decoration
        name: str
        age: int

    model_schema_validator = SchemaValidator(
        {
            'type': 'new-class',
            'class_type': MyCoreModel,
            'schema': {
                'type': 'typed-dict',
                'return_fields_set': True,
                'fields': {'name': {'schema': {'type': 'str'}}, 'age': {'schema': {'type': 'int'}}},
            },
        }
    )
    benchmark(model_schema_validator.validate_python, small_class_data)


@pytest.mark.benchmark(group='string')
def test_core_string_lax(benchmark):
    validator = SchemaValidator({'type': 'str'})
    input_str = 'Hello ' * 20

    benchmark(validator.validate_python, input_str)


@pytest.fixture
def recursive_model_data():
    data = {'width': -1}

    _data = data
    for i in range(100):
        _data['branch'] = {'width': i}
        _data = _data['branch']
    return data


@pytest.mark.skipif(platform.python_implementation() == 'PyPy', reason='crashes on pypy due to recursion depth')
@skip_pydantic
@pytest.mark.benchmark(group='recursive model')
def test_recursive_model_pyd(recursive_model_data, benchmark):
    class PydanticBranch(BaseModel):
        width: int
        branch: Optional['PydanticBranch'] = None  # noqa: F821

    benchmark(PydanticBranch.parse_obj, recursive_model_data)


@pytest.mark.skipif(platform.python_implementation() == 'PyPy', reason='crashes on pypy due to recursion depth')
@pytest.mark.benchmark(group='recursive model')
def test_recursive_model_core(recursive_model_data, benchmark):
    class CoreBranch:
        # this is not required, but it avoids `__fields_set__` being included in `__dict__`
        __slots__ = '__dict__', '__fields_set__'

    v = SchemaValidator(
        {
            'ref': 'Branch',
            'type': 'new-class',
            'class_type': CoreBranch,
            'schema': {
                'type': 'typed-dict',
                'return_fields_set': True,
                'fields': {
                    'width': {'schema': {'type': 'int'}},
                    'branch': {
                        'schema': {'type': 'nullable', 'schema': {'type': 'recursive-ref', 'schema_ref': 'Branch'}},
                        'default': None,
                    },
                },
            },
        }
    )
    benchmark(v.validate_python, recursive_model_data)


@skip_pydantic
@pytest.mark.benchmark(group='List[TypedDict]')
def test_list_of_dict_models_pyd(benchmark):
    class PydanticBranch(BaseModel):
        width: int

    class PydanticRoot(BaseModel):
        __root__: List[PydanticBranch]

    data = [{'width': i} for i in range(100)]
    benchmark(PydanticRoot.parse_obj, data)


@pytest.mark.benchmark(group='List[TypedDict]')
def test_list_of_dict_models_core(benchmark):
    v = SchemaValidator(
        {'type': 'list', 'items_schema': {'type': 'typed-dict', 'fields': {'width': {'schema': {'type': 'int'}}}}}
    )

    data = [{'width': i} for i in range(100)]
    benchmark(v.validate_python, data)


list_of_ints_data = ([i for i in range(1000)], [str(i) for i in range(1000)])


@skip_pydantic
@pytest.mark.benchmark(group='List[int]')
def test_list_of_ints_pyd_py(benchmark):
    class PydanticModel(BaseModel):
        __root__: List[int]

    @benchmark
    def t():
        PydanticModel.parse_obj(list_of_ints_data[0])
        PydanticModel.parse_obj(list_of_ints_data[1])


@pytest.mark.benchmark(group='List[int]')
def test_list_of_ints_core_py(benchmark):
    v = SchemaValidator({'type': 'list', 'items_schema': {'type': 'int'}})

    @benchmark
    def t():
        v.validate_python(list_of_ints_data[0])
        v.validate_python(list_of_ints_data[1])


@skip_pydantic
@pytest.mark.benchmark(group='List[int] JSON')
def test_list_of_ints_pyd_json(benchmark):
    class PydanticModel(BaseModel):
        __root__: List[int]

    json_data = [json.dumps(d) for d in list_of_ints_data]

    @benchmark
    def t():
        PydanticModel.parse_obj(json.loads(json_data[0]))
        PydanticModel.parse_obj(json.loads(json_data[1]))


@pytest.mark.benchmark(group='List[int] JSON')
def test_list_of_ints_core_json(benchmark):
    v = SchemaValidator({'type': 'list', 'items_schema': {'type': 'int'}})

    json_data = [json.dumps(d) for d in list_of_ints_data]

    @benchmark
    def t():
        v.validate_json(json_data[0])
        v.validate_json(json_data[1])


@skip_pydantic
@pytest.mark.benchmark(group='List[Any]')
def test_list_of_any_pyd_py(benchmark):
    class PydanticModel(BaseModel):
        __root__: list

    @benchmark
    def t():
        PydanticModel.parse_obj(list_of_ints_data[0])
        PydanticModel.parse_obj(list_of_ints_data[1])


@pytest.mark.benchmark(group='List[Any]')
def test_list_of_any_core_py(benchmark):
    v = SchemaValidator({'type': 'list'})

    @benchmark
    def t():
        v.validate_python(list_of_ints_data[0])
        v.validate_python(list_of_ints_data[1])


set_of_ints_data = ({i for i in range(1000)}, {str(i) for i in range(1000)})


@skip_pydantic
@pytest.mark.benchmark(group='Set[int]')
def test_set_of_ints_pyd(benchmark):
    class PydanticModel(BaseModel):
        __root__: Set[int]

    @benchmark
    def t():
        PydanticModel.parse_obj(set_of_ints_data[0])
        PydanticModel.parse_obj(set_of_ints_data[1])


@pytest.mark.benchmark(group='Set[int]')
def test_set_of_ints_core(benchmark):
    v = SchemaValidator({'type': 'set', 'items_schema': {'type': 'int'}})

    @benchmark
    def t():
        v.validate_python(set_of_ints_data[0])
        v.validate_python(set_of_ints_data[1])


@skip_pydantic
@pytest.mark.benchmark(group='Set[int] JSON')
def test_set_of_ints_pyd_json(benchmark):
    class PydanticModel(BaseModel):
        __root__: Set[int]

    json_data = [json.dumps(list(d)) for d in set_of_ints_data]

    @benchmark
    def t():
        PydanticModel.parse_obj(json.loads(json_data[0]))
        PydanticModel.parse_obj(json.loads(json_data[1]))


@pytest.mark.benchmark(group='Set[int] JSON')
def test_set_of_ints_core_json(benchmark):
    v = SchemaValidator({'type': 'set', 'items_schema': {'type': 'int'}})

    json_data = [json.dumps(list(d)) for d in set_of_ints_data]

    @benchmark
    def t():
        v.validate_json(json_data[0])
        v.validate_json(json_data[1])


frozenset_of_ints = frozenset({i for i in range(1000)})


@skip_pydantic
@pytest.mark.benchmark(group='FrozenSet[int]')
def test_frozenset_of_ints_pyd(benchmark):
    class PydanticModel(BaseModel):
        __root__: FrozenSet[int]

    benchmark(PydanticModel.parse_obj, frozenset_of_ints)


@pytest.mark.benchmark(group='FrozenSet[int]')
def test_frozenset_of_ints_core(benchmark):
    v = SchemaValidator({'type': 'frozenset', 'items_schema': {'type': 'int'}})

    benchmark(v.validate_python, frozenset_of_ints)


dict_of_ints_data = ({str(i): i for i in range(1000)}, {str(i): str(i) for i in range(1000)})


@skip_pydantic
@pytest.mark.benchmark(group='Dict[str, int]')
def test_dict_of_ints_pyd(benchmark):
    class PydanticModel(BaseModel):
        __root__: Dict[str, int]

    @benchmark
    def t():
        PydanticModel.parse_obj(dict_of_ints_data[0])
        PydanticModel.parse_obj(dict_of_ints_data[1])


@pytest.mark.benchmark(group='Dict[str, int]')
def test_dict_of_ints_core(benchmark):
    v = SchemaValidator({'type': 'dict', 'keys_schema': 'str', 'values_schema': 'int'})

    @benchmark
    def t():
        v.validate_python(dict_of_ints_data[0])
        v.validate_python(dict_of_ints_data[1])


@pytest.mark.benchmark(group='Dict[any, any]')
def test_dict_of_any_core(benchmark):
    v = SchemaValidator({'type': 'dict'})

    @benchmark
    def t():
        v.validate_python(dict_of_ints_data[0])
        v.validate_python(dict_of_ints_data[1])


@skip_pydantic
@pytest.mark.benchmark(group='Dict[str, int] JSON')
def test_dict_of_ints_pyd_json(benchmark):
    class PydanticModel(BaseModel):
        __root__: Dict[str, int]

    json_data = [json.dumps(d) for d in dict_of_ints_data]

    @benchmark
    def t():
        PydanticModel.parse_obj(json.loads(json_data[0]))
        PydanticModel.parse_obj(json.loads(json_data[1]))


@pytest.mark.benchmark(group='Dict[str, int] JSON')
def test_dict_of_ints_core_json(benchmark):
    v = SchemaValidator({'type': 'dict', 'keys_schema': 'str', 'values_schema': 'int'})

    json_data = [json.dumps(d) for d in dict_of_ints_data]

    @benchmark
    def t():
        v.validate_json(json_data[0])
        v.validate_json(json_data[1])


many_models_data = [{'age': i} for i in range(1000)]


@skip_pydantic
@pytest.mark.benchmark(group='List[SimpleMode]')
def test_many_models_pyd(benchmark):
    class SimpleMode(BaseModel):
        age: int

    class PydanticModel(BaseModel):
        __root__: List[SimpleMode]

    benchmark(PydanticModel.parse_obj, many_models_data)


@pytest.mark.benchmark(group='List[DictSimpleMode]')
def test_many_models_core_dict(benchmark):
    model_schema = {'type': 'list', 'items_schema': {'type': 'typed-dict', 'fields': {'age': {'schema': 'int'}}}}
    v = SchemaValidator(model_schema)
    benchmark(v.validate_python, many_models_data)


@pytest.mark.benchmark(group='List[SimpleMode]')
def test_many_models_core_model(benchmark):
    class MyCoreModel:
        __slots__ = '__dict__', '__fields_set__'

    v = SchemaValidator(
        {
            'type': 'list',
            'items_schema': {
                'type': 'new-class',
                'class_type': MyCoreModel,
                'schema': {'type': 'typed-dict', 'return_fields_set': True, 'fields': {'age': {'schema': 'int'}}},
            },
        }
    )
    benchmark(v.validate_python, many_models_data)


list_of_nullable_data = [None if i % 2 else i for i in range(1000)]


@skip_pydantic
@pytest.mark.benchmark(group='List[Nullable[int]]')
def test_list_of_nullable_pyd(benchmark):
    class PydanticModel(BaseModel):
        __root__: List[Optional[int]]

    benchmark(PydanticModel.parse_obj, list_of_nullable_data)


@pytest.mark.benchmark(group='List[Nullable[int]]')
def test_list_of_nullable_core(benchmark):
    v = SchemaValidator({'type': 'list', 'items_schema': {'type': 'nullable', 'schema': 'int'}})

    benchmark(v.validate_python, list_of_nullable_data)


some_bytes = b'0' * 1000


@pytest.mark.benchmark(group='bytes')
def test_bytes_core(benchmark):
    v = SchemaValidator({'type': 'bytes'})

    benchmark(v.validate_python, some_bytes)


@skip_pydantic
@pytest.mark.benchmark(group='bytes')
def test_bytes_pyd(benchmark):
    class PydanticModel(BaseModel):
        __root__: bytes

    benchmark(PydanticModel.parse_obj, some_bytes)


class TestBenchmarkDateTime:
    @pytest.fixture(scope='class')
    def pydantic_model(self):
        class PydanticModel(BaseModel):
            dt: datetime

        return PydanticModel

    @pytest.fixture(scope='class')
    def core_validator(self):
        class CoreModel:
            __slots__ = '__dict__', '__fields_set__'

        return SchemaValidator(
            {
                'type': 'new-class',
                'class_type': CoreModel,
                'schema': {'type': 'typed-dict', 'return_fields_set': True, 'fields': {'dt': {'schema': 'datetime'}}},
            }
        )

    @pytest.fixture(scope='class')
    def datetime_raw(self):
        return datetime.utcnow().replace(tzinfo=timezone.utc) + timedelta(days=1)

    @pytest.fixture(scope='class')
    def datetime_str(self, datetime_raw):
        return str(datetime_raw)

    @pytest.fixture(scope='class')
    def python_data_dict(self, datetime_raw):
        return {'dt': datetime_raw}

    @pytest.fixture(scope='class')
    def json_dict_data(self, datetime_str):
        return json.dumps({'dt': datetime_str})

    @skip_pydantic
    @pytest.mark.benchmark(group='datetime model - python')
    def test_pyd_python(self, pydantic_model, benchmark, python_data_dict):
        benchmark(pydantic_model.parse_obj, python_data_dict)

    @pytest.mark.benchmark(group='datetime model - python')
    def test_core_python(self, core_validator, benchmark, python_data_dict):
        benchmark(core_validator.validate_python, python_data_dict)

    @skip_pydantic
    @pytest.mark.benchmark(group='datetime model - JSON')
    def test_model_pyd_json(self, pydantic_model, benchmark, json_dict_data):
        @benchmark
        def pydantic_json():
            obj = json.loads(json_dict_data)
            return pydantic_model.parse_obj(obj)

    @pytest.mark.benchmark(group='datetime model - JSON')
    def test_model_core_json(self, core_validator, benchmark, json_dict_data):
        benchmark(core_validator.validate_json, json_dict_data)

    @pytest.mark.benchmark(group='datetime datetime')
    def test_core_raw(self, benchmark, datetime_raw):
        v = SchemaValidator({'type': 'datetime'})

        benchmark(v.validate_python, datetime_raw)

    @pytest.mark.benchmark(group='datetime str')
    def test_core_str(self, benchmark, datetime_str):
        v = SchemaValidator({'type': 'datetime'})

        benchmark(v.validate_python, datetime_str)

    @pytest.mark.benchmark(group='datetime future')
    def test_core_future(self, benchmark, datetime_raw):
        v = SchemaValidator({'type': 'datetime', 'gt': datetime.now()})

        benchmark(v.validate_python, datetime_raw)

    @pytest.mark.benchmark(group='datetime future')
    def test_core_future_str(self, benchmark, datetime_str):
        v = SchemaValidator({'type': 'datetime', 'gt': datetime.now()})

        benchmark(v.validate_python, datetime_str)


class TestBenchmarkDateX:
    @pytest.fixture(scope='class')
    def validator(self):
        return SchemaValidator({'type': 'date'})

    @pytest.mark.benchmark(group='date from date')
    def test_date_from_date(self, benchmark, validator):
        benchmark(validator.validate_python, date.today())

    @pytest.mark.benchmark(group='date from str')
    def test_date_from_str(self, benchmark, validator):
        benchmark(validator.validate_python, str(date.today()))

    @pytest.mark.benchmark(group='date from datetime')
    def test_date_from_datetime(self, benchmark, validator):
        benchmark(validator.validate_python, datetime(2000, 1, 1))

    @pytest.mark.benchmark(group='date from datetime str')
    def test_date_from_datetime_str(self, benchmark, validator):
        benchmark(validator.validate_python, str(datetime(2000, 1, 1)))

    @pytest.mark.benchmark(group='date future')
    def test_core_future(self, benchmark):
        v = SchemaValidator({'type': 'date', 'gt': date.today()})

        benchmark(v.validate_python, date(2032, 1, 1))

    @pytest.mark.benchmark(group='date future')
    def test_core_future_str(self, benchmark):
        v = SchemaValidator({'type': 'date', 'gt': date.today()})

        benchmark(v.validate_python, str(date(2032, 1, 1)))


class TestBenchmarkUnion:
    @pytest.mark.benchmark(group='smart-union')
    def test_smart_union_core(self, benchmark):
        v = SchemaValidator({'type': 'union', 'choices': [{'type': 'bool'}, {'type': 'int'}, {'type': 'str'}]})

        benchmark(v.validate_python, 1)

    @skip_pydantic
    @pytest.mark.benchmark(group='smart-union')
    def test_smart_union_pyd(self, benchmark):
        # default pydantic-core behavior matches pydantic one with `Config.smart_union`
        class PydanticModel(BaseModel, smart_union=True):
            __root__: Union[bool, int, str]

        benchmark(PydanticModel.parse_obj, 1)

    @pytest.mark.benchmark(group='smart-union-coerce')
    def test_smart_union_coerce_core(self, benchmark):
        v = SchemaValidator({'type': 'union', 'choices': [{'type': 'bool'}, {'type': 'str'}]})

        benchmark(v.validate_python, 1)  # will be True

    @skip_pydantic
    @pytest.mark.benchmark(group='smart-union-coerce')
    def test_smart_union_coerce_pyd(self, benchmark):
        class PydanticModel(BaseModel, smart_union=True):
            __root__: Union[bool, str]

        benchmark(PydanticModel.parse_obj, 1)  # will be True

    @pytest.mark.benchmark(group='strict-union')
    def test_strict_union_core(self, benchmark):
        v = SchemaValidator(
            {'type': 'union', 'strict': True, 'choices': [{'type': 'bool'}, {'type': 'int'}, {'type': 'str'}]}
        )

        benchmark(v.validate_python, 1)

    @skip_pydantic
    @pytest.mark.benchmark(group='strict-union')
    def test_strict_union_pyd(self, benchmark):
        class PydanticModel(BaseModel):
            __root__: Union[StrictBool, StrictInt, StrictStr]

        benchmark(PydanticModel.parse_obj, 1)  # will be True

    @pytest.mark.benchmark(group='strict-union-error')
    def test_strict_union_error_core(self, benchmark):
        v = SchemaValidator({'type': 'union', 'strict': True, 'choices': [{'type': 'bool'}, {'type': 'str'}]})

        def validate_with_expected_error():
            try:
                v.validate_python(2)
                assert False
            except CoreValidationError:
                assert True

        benchmark(validate_with_expected_error)

    @skip_pydantic
    @pytest.mark.benchmark(group='strict-union-error')
    def test_strict_union_error_pyd(self, benchmark):
        class PydanticModel(BaseModel):
            __root__: Union[StrictBool, StrictStr]

        def validate_with_expected_error():
            try:
                PydanticModel.parse_obj(2)
                assert False
            except PydanticValidationError:
                assert True

        benchmark(validate_with_expected_error)


@pytest.mark.benchmark(group='raise-error')
def test_dont_raise_error(benchmark):
    def f(input_value, **kwargs):
        return input_value

    v = SchemaValidator({'type': 'function', 'mode': 'plain', 'function': f})

    @benchmark
    def t():
        v.validate_python(42)


@pytest.mark.benchmark(group='raise-error')
def test_raise_error_value_error(benchmark):
    def f(input_value, **kwargs):
        raise ValueError('this is a custom error')

    v = SchemaValidator({'type': 'function', 'mode': 'plain', 'function': f})

    @benchmark
    def t():
        try:
            v.validate_python(42)
        except ValidationError:
            pass
        else:
            raise RuntimeError('expected ValidationError')


@pytest.mark.benchmark(group='raise-error')
def test_raise_error_custom(benchmark):
    def f(input_value, **kwargs):
        raise PydanticValueError('my_error', 'this is a custom error {foo}', {'foo': 'FOOBAR'})

    v = SchemaValidator({'type': 'function', 'mode': 'plain', 'function': f})

    @benchmark
    def t():
        try:
            v.validate_python(42)
        except ValidationError:
            pass
        else:
            raise RuntimeError('expected ValidationError')


@pytest.mark.benchmark(group='tuple')
def test_positional_tuple(benchmark):
    v = SchemaValidator({'type': 'tuple', 'mode': 'positional', 'items_schema': ['int', 'int', 'int', 'int', 'int']})
    assert v.validate_python((1, 2, 3, '4', 5)) == (1, 2, 3, 4, 5)

    benchmark(v.validate_python, (1, 2, 3, '4', 5))


@pytest.mark.benchmark(group='tuple')
def test_variable_tuple(benchmark):
    v = SchemaValidator({'type': 'tuple', 'items_schema': 'int'})
    assert v.validate_python((1, 2, 3, '4', 5)) == (1, 2, 3, 4, 5)

    benchmark(v.validate_python, (1, 2, 3, '4', 5))


@pytest.mark.benchmark(group='tuple-many')
def test_tuple_many_variable(benchmark):
    v = SchemaValidator({'type': 'tuple', 'items_schema': 'int'})
    assert v.validate_python(list(range(10))) == tuple(range(10))

    benchmark(v.validate_python, list(range(10)))


@pytest.mark.benchmark(group='tuple-many')
def test_tuple_many_positional(benchmark):
    v = SchemaValidator({'type': 'tuple', 'mode': 'positional', 'items_schema': [], 'extra_schema': 'int'})
    assert v.validate_python(list(range(10))) == tuple(range(10))

    benchmark(v.validate_python, list(range(10)))


@pytest.mark.benchmark(group='arguments')
def test_arguments(benchmark):
    v = SchemaValidator(
        {
            'type': 'arguments',
            'arguments_schema': [
                {'name': 'args1', 'mode': 'positional_only', 'schema': 'int'},
                {'name': 'args2', 'mode': 'positional_only', 'schema': 'str'},
                {'name': 'a', 'mode': 'positional_or_keyword', 'schema': 'bool'},
                {'name': 'b', 'mode': 'keyword_only', 'schema': 'str'},
                {'name': 'c', 'mode': 'keyword_only', 'schema': 'int'},
            ],
        }
    )
    assert v.validate_python(((1, 'a', 'true'), {'b': 'bb', 'c': 3})) == ((1, 'a', True), {'b': 'bb', 'c': 3})

    benchmark(v.validate_python, ((1, 'a', 'true'), {'b': 'bb', 'c': 3}))



from datetime import date, datetime, time

from pydantic_core import Schema, SchemaError, SchemaValidator


class Foo:
    bar: str


def foo(bar: str) -> None:
    ...


def test_schema_typing() -> None:
    # this gets run by pyright, but we also check that it executes
    schema: Schema = {'type': 'union', 'choices': ['int', {'type': 'int', 'ge': 1}, {'type': 'float', 'lt': 1.0}]}
    SchemaValidator(schema)
    schema: Schema = {
        'type': 'tagged-union',
        'discriminator': 'kind',
        'choices': {
            'apple': {'type': 'typed-dict', 'fields': {'pips': {'schema': {'type': 'int'}}}},
            'banana': {'type': 'typed-dict', 'fields': {'curvature': {'schema': {'type': 'float'}}}},
        },
    }
    SchemaValidator(schema)
    schema: Schema = {'type': 'int', 'ge': 1}
    SchemaValidator(schema)
    schema: Schema = {'type': 'float', 'lt': 1.0}
    SchemaValidator(schema)
    schema: Schema = {'type': 'str', 'pattern': r'http://.*'}
    SchemaValidator(schema)
    schema: Schema = {'type': 'bool', 'strict': False}
    SchemaValidator(schema)
    schema: Schema = {'type': 'literal', 'expected': [1, '1']}
    SchemaValidator(schema)
    schema: Schema = {'type': 'any'}
    SchemaValidator(schema)
    schema: Schema = {'type': 'none'}
    SchemaValidator(schema)
    schema: Schema = {'type': 'bytes'}
    SchemaValidator(schema)
    schema: Schema = {'type': 'list', 'items_schema': {'type': 'str'}, 'min_items': 3}
    SchemaValidator(schema)
    schema: Schema = {'type': 'set', 'items_schema': {'type': 'str'}, 'max_items': 3}
    SchemaValidator(schema)
    schema: Schema = {'type': 'tuple', 'mode': 'variable', 'items_schema': {'type': 'str'}, 'max_items': 3}
    SchemaValidator(schema)
    schema: Schema = {'type': 'tuple', 'mode': 'positional', 'items_schema': [{'type': 'str'}, {'type': 'int'}]}
    SchemaValidator(schema)
    schema: Schema = {'type': 'frozenset', 'items_schema': {'type': 'str'}, 'max_items': 3}
    SchemaValidator(schema)
    schema: Schema = {'type': 'dict', 'keys_schema': {'type': 'str'}, 'values_schema': {'type': 'any'}}
    SchemaValidator(schema)
    schema: Schema = {
        'type': 'new-class',
        'class_type': Foo,
        'schema': {'type': 'typed-dict', 'return_fields_set': True, 'fields': {'bar': {'schema': {'type': 'str'}}}},
    }
    SchemaValidator(schema)
    schema: Schema = {
        'type': 'typed-dict',
        'fields': {
            'a': {'schema': {'type': 'str'}},
            'b': {'schema': {'type': 'str'}, 'alias': 'foobar'},
            'c': {'schema': {'type': 'str'}, 'alias': [['foobar', 0, 'bar'], ['foo']]},
            'd': {'schema': {'type': 'str'}, 'default': 'spam'},
        },
    }
    SchemaValidator(schema)
    schema: Schema = {'type': 'function', 'mode': 'wrap', 'function': foo, 'schema': {'type': 'str'}}
    SchemaValidator(schema)
    schema: Schema = {'type': 'function', 'mode': 'plain', 'function': foo}
    SchemaValidator(schema)
    schema: Schema = {
        'ref': 'Branch',
        'type': 'typed-dict',
        'fields': {
            'name': {'schema': {'type': 'str'}},
            'sub_branch': {
                'schema': {
                    'type': 'union',
                    'choices': [{'type': 'none'}, {'type': 'recursive-ref', 'schema_ref': 'Branch'}],
                },
                'default': None,
            },
        },
    }
    SchemaValidator(schema)
    schema: Schema = {'type': 'date', 'le': date.today()}
    SchemaValidator(schema)
    schema: Schema = {'type': 'time', 'lt': time(12, 13, 14)}
    SchemaValidator(schema)
    schema: Schema = {'type': 'datetime', 'ge': datetime.now()}
    SchemaValidator(schema)
    schema: Schema = {'type': 'is-instance', 'class_': Foo}
    SchemaValidator(schema)
    schema: Schema = {'type': 'callable'}
    SchemaValidator(schema)

    schema: Schema = {
        'type': 'arguments',
        'arguments_schema': [
            {'name': 'a', 'mode': 'positional_only', 'schema': 'int'},
            {'name': 'b', 'schema': 'str'},
            {'name': 'c', 'mode': 'keyword_only', 'schema': 'bool'},
        ],
    }
    SchemaValidator(schema)

    schema: Schema = {'type': 'call', 'arguments_schema': 'any', 'function': foo}
    SchemaValidator(schema)


def test_schema_typing_error() -> None:
    _: Schema = {'type': 'wrong'}  # type: ignore


def test_schema_validator() -> None:
    SchemaValidator('int')


def test_schema_validator_wrong() -> None:
    # use this instead of pytest.raises since pyright complains about input when pytest isn't installed
    try:
        SchemaValidator('bad')  # type: ignore
    except SchemaError:
        pass
    else:
        raise AssertionError('SchemaValidator did not raise SchemaError')



import functools
import importlib.util
import json
import os
import re
from dataclasses import dataclass
from pathlib import Path
from typing import Any, Type

import pytest
from hypothesis import settings
from typing_extensions import Literal

from pydantic_core import SchemaValidator

__all__ = 'Err', 'PyAndJson', 'plain_repr'

hyp_max_examples = os.getenv('HYPOTHESIS_MAX_EXAMPLES')
if hyp_max_examples:
    settings.register_profile('custom', max_examples=int(hyp_max_examples))
    settings.load_profile('custom')


def plain_repr(obj):
    r = repr(obj)
    r = re.sub(r',\s*([)}])', r'\1', r)
    r = re.sub(r'\s+', '', r)
    return r


@dataclass
class Err:
    message: str
    errors: Any = None

    def __repr__(self):
        if self.errors:
            return f'Err({self.message!r}, errors={self.errors!r})'
        else:
            return f'Err({self.message!r})'


class PyAndJsonValidator:
    def __init__(self, schema, validator_type: 'Literal["json", "python"] | None' = None):
        self.validator = SchemaValidator(schema)
        self.validator_type = validator_type

    def validate_python(self, py_input, strict: 'bool | None' = None):
        return self.validator.validate_python(py_input, strict)

    def validate_test(self, py_input, strict: 'bool | None' = None):
        if self.validator_type == 'json':
            return self.validator.validate_json(json.dumps(py_input), strict)
        elif self.validator_type == 'python':
            return self.validator.validate_python(py_input, strict)

    def isinstance_test(self, py_input, strict: 'bool | None' = None):
        if self.validator_type == 'json':
            return self.validator.isinstance_json(json.dumps(py_input), strict)
        elif self.validator_type == 'python':
            return self.validator.isinstance_python(py_input, strict)


PyAndJson = Type[PyAndJsonValidator]


@pytest.fixture(params=['python', 'json'])
def py_and_json(request) -> PyAndJson:
    class ChosenPyAndJsonValidator(PyAndJsonValidator):
        __init__ = functools.partialmethod(PyAndJsonValidator.__init__, validator_type=request.param)

    return ChosenPyAndJsonValidator


@pytest.fixture
def tmp_work_path(tmp_path: Path):
    """
    Create a temporary working directory.
    """
    previous_cwd = Path.cwd()
    os.chdir(tmp_path)

    yield tmp_path

    os.chdir(previous_cwd)


@pytest.fixture
def import_execute(request, tmp_work_path: Path):
    def _import_execute(source: str, *, custom_module_name: 'str | None' = None):
        example_bash_file = tmp_work_path / 'example.sh'
        example_bash_file.write_text('#!/bin/sh\necho testing')
        example_bash_file.chmod(0o755)
        (tmp_work_path / 'first/path').mkdir(parents=True, exist_ok=True)
        (tmp_work_path / 'second/path').mkdir(parents=True, exist_ok=True)

        module_name = custom_module_name or request.node.name

        module_path = tmp_work_path / f'{module_name}.py'
        module_path.write_text(source)
        spec = importlib.util.spec_from_file_location('__main__', str(module_path))
        module = importlib.util.module_from_spec(spec)
        try:
            spec.loader.exec_module(module)
        except KeyboardInterrupt:
            print('KeyboardInterrupt')
        else:
            return module

    return _import_execute



import pytest

from pydantic_core import SchemaValidator, ValidationError


def test_isinstance():
    v = SchemaValidator({'type': 'int'})
    assert v.validate_python(123) == 123
    assert v.isinstance_python(123) is True
    assert v.validate_python('123') == 123
    assert v.isinstance_python('123') is True

    with pytest.raises(ValidationError, match='Input should be a valid integer'):
        v.validate_python('foo')

    assert v.isinstance_python('foo') is False


def test_isinstance_strict():
    v = SchemaValidator({'type': 'int', 'strict': True})
    assert v.validate_python(123) == 123
    assert v.isinstance_python(123) is True

    with pytest.raises(ValidationError, match='Input should be a valid integer'):
        v.validate_python('123')

    assert v.isinstance_python('123') is False


def test_isinstance_json():
    v = SchemaValidator({'type': 'int'})
    assert v.validate_json('123') == 123
    assert v.isinstance_json('123') is True
    assert v.validate_json('"123"') == 123
    assert v.isinstance_json('"123"') is True

    with pytest.raises(ValidationError, match='Input should be a valid integer'):
        v.validate_json('"foo"')

    assert v.isinstance_json('"foo"') is False

    with pytest.raises(ValidationError, match=r'Invalid JSON: expected value at line 1 column 1 \[kind=invalid_json,'):
        v.validate_json('x')

    # invalid json returns False, not an error!
    assert v.isinstance_json('x') is False


def test_internal_error():
    v = SchemaValidator(
        {
            'type': 'new-class',
            'class_type': int,
            'schema': {'type': 'typed-dict', 'return_fields_set': True, 'fields': {'f': {'schema': 'int'}}},
        }
    )
    with pytest.raises(AttributeError, match="'int' object has no attribute '__dict__'"):
        v.validate_python({'f': 123})

    with pytest.raises(AttributeError, match="'int' object has no attribute '__dict__'"):
        v.validate_json('{"f": 123}')

    with pytest.raises(AttributeError, match="'int' object has no attribute '__dict__'"):
        v.isinstance_python({'f': 123})

    with pytest.raises(AttributeError, match="'int' object has no attribute '__dict__'"):
        v.isinstance_json('{"f": 123}')



import pytest

from pydantic_core import SchemaError, SchemaValidator, ValidationError


class Foo:
    pass


class Bar(Foo):
    pass


class Spam:
    pass


def test_is_instance():
    v = SchemaValidator({'type': 'is-instance', 'class_': Foo})
    foo = Foo()
    assert v.validate_python(foo) == foo
    assert v.isinstance_python(foo) is True
    bar = Bar()
    assert v.validate_python(bar) == bar
    s = Spam()
    assert v.isinstance_python(s) is False
    with pytest.raises(ValidationError) as exc_info:
        v.validate_python(s)

    assert exc_info.value.errors() == [
        {
            'kind': 'is_instance_of',
            'loc': [],
            'message': 'Input should be an instance of Foo',
            'input_value': s,
            'context': {'class': 'Foo'},
        }
    ]
    with pytest.raises(ValidationError, match='kind=is_instance_of'):
        v.validate_python(Foo)

    with pytest.raises(ValidationError, match='kind=is_instance_of'):
        v.validate_json('"foo"')


@pytest.mark.parametrize(
    'schema_class,input_val,value',
    [
        (Foo, Foo(), True),
        (Foo, Foo, False),
        (Foo, Bar(), True),
        (Foo, Bar, False),
        (Bar, Foo(), False),
        (Bar, Foo, False),
        (dict, {1: 2}, True),
        (dict, {1, 2}, False),
        (type, Foo, True),
        (type, Foo(), False),
    ],
)
def test_is_instance_cases(schema_class, input_val, value):
    v = SchemaValidator({'type': 'is-instance', 'class_': schema_class})
    assert v.isinstance_python(input_val) == value


@pytest.mark.parametrize('input_cls', [123, 'foo', Foo(), [], {1: 2}])
def test_is_instance_invalid(input_cls):
    with pytest.raises(SchemaError, match="object cannot be converted to 'PyType'"):
        SchemaValidator({'type': 'is-instance', 'class_': input_cls})


class HasIsInstanceMeta(type):
    def __instancecheck__(self, instance) -> bool:
        if 'error' in repr(instance):
            # an error here comes from a problem in the schema, not in the input value, so raise as internal error
            raise TypeError('intentional error')
        return 'true' in repr(instance)


class HasIsInstance(metaclass=HasIsInstanceMeta):
    pass


def test_instancecheck():
    v = SchemaValidator({'type': 'is-instance', 'class_': HasIsInstance})
    assert v.validate_python('true') == 'true'

    with pytest.raises(ValidationError, match='kind=is_instance_of'):
        v.validate_python('other')

    with pytest.raises(TypeError, match='intentional error'):
        v.validate_python('error')


def test_repr():
    v = SchemaValidator({'type': 'union', 'choices': ['int', {'type': 'is-instance', 'class_': Foo}]})
    assert v.isinstance_python(4) is True
    assert v.isinstance_python(Bar()) is True
    assert v.isinstance_python('foo') is False

    with pytest.raises(ValidationError, match=r'is-instance\[Foo\]\s+Input should be an instance of Foo'):
        v.validate_python('foo')



import pytest
from dirty_equals import HasAttributes, IsInstance

from pydantic_core import SchemaValidator, ValidationError

from .conftest import plain_repr


def test_on_field():
    v = SchemaValidator({'type': 'str', 'min_length': 2, 'max_length': 5})
    r = plain_repr(v)
    assert 'min_length:Some(2)' in r
    assert 'max_length:Some(5)' in r
    assert v.isinstance_python('test') is True
    assert v.isinstance_python('test long') is False


def test_on_config():
    v = SchemaValidator({'type': 'str'}, {'str_max_length': 5})
    assert 'max_length:Some(5)' in plain_repr(v)
    assert v.isinstance_python('test') is True
    assert v.isinstance_python('test long') is False


def test_field_priority_arg():
    v = SchemaValidator({'type': 'str', 'max_length': 5}, {'str_max_length': 10})
    assert 'max_length:Some(5)' in plain_repr(v)
    assert v.isinstance_python('test') is True
    assert v.isinstance_python('test long') is False


class MyModel:
    # this is not required, but it avoids `__fields_set__` being included in `__dict__`
    __slots__ = '__dict__', '__fields_set__'


def test_on_model_class():

    v = SchemaValidator(
        {
            'type': 'new-class',
            'class_type': MyModel,
            'config': {'str_max_length': 5},
            'schema': {'type': 'typed-dict', 'return_fields_set': True, 'fields': {'f': {'schema': 'str'}}},
        }
    )
    assert 'max_length:Some(5)' in plain_repr(v)
    assert v.isinstance_python({'f': 'test'}) is True
    assert v.isinstance_python({'f': 'test long'}) is False


def test_field_priority_model():
    v = SchemaValidator(
        {
            'type': 'new-class',
            'class_type': MyModel,
            'config': {'str_max_length': 10},
            'schema': {
                'type': 'typed-dict',
                'return_fields_set': True,
                'fields': {'f': {'schema': {'type': 'str', 'max_length': 5}}},
            },
        }
    )
    assert 'max_length:Some(5)' in plain_repr(v)
    assert v.isinstance_python({'f': 'test'}) is True
    assert v.isinstance_python({'f': 'test long'}) is False


def test_parent_priority():
    v = SchemaValidator(
        {
            'type': 'new-class',
            'class_type': MyModel,
            'schema': {'type': 'typed-dict', 'return_fields_set': True, 'fields': {'f': {'schema': 'str'}}},
            'config': {'str_min_length': 2, 'str_max_length': 10},
        },
        {'str_max_length': 5, 'config_choose_priority': 1},
    )
    r = plain_repr(v)
    assert 'min_length:Some(2)' not in r  # child is ignored
    assert 'max_length:Some(5)' in r
    assert v.isinstance_python({'f': 'test'}) is True
    assert v.isinstance_python({'f': 't'}) is True
    assert v.isinstance_python({'f': 'test long'}) is False


def test_child_priority():
    v = SchemaValidator(
        {
            'type': 'new-class',
            'class_type': MyModel,
            'schema': {'type': 'typed-dict', 'return_fields_set': True, 'fields': {'f': {'schema': 'str'}}},
            'config': {'str_max_length': 5, 'config_choose_priority': 1},
        },
        {'str_min_length': 2, 'str_max_length': 10},
    )
    r = plain_repr(v)
    assert 'min_length:Some(2)' not in r  # parent is ignored
    assert 'max_length:Some(5)' in r
    assert v.isinstance_python({'f': 'test'}) is True
    assert v.isinstance_python({'f': 't'}) is True
    assert v.isinstance_python({'f': 'test long'}) is False


def test_merge_child_wins():
    v = SchemaValidator(
        {
            'type': 'new-class',
            'class_type': MyModel,
            'schema': {'type': 'typed-dict', 'return_fields_set': True, 'fields': {'f': {'schema': 'str'}}},
            'config': {'str_max_length': 5},
        },
        {'str_min_length': 2, 'str_max_length': 10},
    )
    r = plain_repr(v)
    assert 'min_length:Some(2)' in r
    assert 'max_length:Some(5)' in r
    assert v.isinstance_python({'f': 'test'}) is True
    assert v.isinstance_python({'f': 't'}) is False
    assert v.isinstance_python({'f': 'test long'}) is False


def test_merge_parent_wins():
    v = SchemaValidator(
        {
            'type': 'new-class',
            'class_type': MyModel,
            'schema': {'type': 'typed-dict', 'return_fields_set': True, 'fields': {'f': {'schema': 'str'}}},
            'config': {'str_max_length': 5},
        },
        {'str_min_length': 2, 'str_max_length': 10, 'config_merge_priority': 1},
    )
    r = plain_repr(v)
    assert 'min_length:Some(2)' in r
    assert 'max_length:Some(10)' in r
    assert 'max_length:Some(5)' not in r
    assert v.isinstance_python({'f': 'test'}) is True
    assert v.isinstance_python({'f': 't'}) is False
    assert v.isinstance_python({'f': 'test long'}) is True
    assert v.isinstance_python({'f': 'test very long'}) is False


def test_sub_model_merge():
    v = SchemaValidator(
        {
            'type': 'new-class',
            'class_type': MyModel,
            'schema': {
                'type': 'typed-dict',
                'return_fields_set': True,
                'fields': {
                    'f': {'schema': 'str'},
                    'sub_model': {
                        'schema': {
                            'type': 'new-class',
                            'class_type': MyModel,
                            'schema': {
                                'type': 'typed-dict',
                                'return_fields_set': True,
                                'fields': {'f': {'schema': 'str'}},
                            },
                            'config': {'str_max_length': 6, 'str_to_upper': True},
                        }
                    },
                },
            },
            'config': {'str_min_length': 1, 'str_max_length': 4},
        }
    )
    # f should have bounds 1-4
    # sub_model.f should have bounds 1-6, and should be upper-cased
    output = v.validate_python({'f': 'test', 'sub_model': {'f': 'tests'}})
    assert output == IsInstance(MyModel) & HasAttributes(f='test', sub_model=HasAttributes(f='TESTS'))
    with pytest.raises(ValidationError) as exc_info:
        v.validate_python({'f': 'tests', 'sub_model': {'f': ''}})
    assert exc_info.value.errors() == [
        {
            'kind': 'too_long',
            'loc': ['f'],
            'message': 'String should have at most 4 characters',
            'input_value': 'tests',
            'context': {'max_length': 4},
        },
        {
            'kind': 'too_short',
            'loc': ['sub_model', 'f'],
            'message': 'String should have at least 1 characters',
            'input_value': '',
            'context': {'min_length': 1},
        },
    ]



import sys
from typing import Any, TypedDict

from pydantic_core._types import Config, Schema

if sys.version_info < (3, 11):
    from typing_extensions import NotRequired
else:
    from typing import NotRequired

__all__ = '__version__', 'SchemaValidator', 'SchemaError', 'ValidationError', 'PydanticValueError'
__version__: str

class SchemaValidator:
    def __init__(self, schema: Schema, config: 'Config | None' = None) -> None: ...
    def validate_python(self, input: Any, strict: 'bool | None' = None, context: Any = None) -> Any: ...
    def isinstance_python(self, input: Any, strict: 'bool | None' = None, context: Any = None) -> bool: ...
    def validate_json(
        self, input: 'str | bytes | bytearray', strict: 'bool | None' = None, context: Any = None
    ) -> Any: ...
    def isinstance_json(
        self, input: 'str | bytes | bytearray', strict: 'bool | None' = None, context: Any = None
    ) -> bool: ...
    def validate_assignment(self, field: str, input: Any, data: 'dict[str, Any]') -> 'dict[str, Any]': ...

class SchemaError(Exception):
    pass

class ErrorDetails(TypedDict):
    kind: str
    loc: 'list[int | str]'
    message: str
    input_value: Any
    context: NotRequired['dict[str, Any]']

class ValidationError(ValueError):
    title: str

    def error_count(self) -> int: ...
    def errors(self) -> 'list[ErrorDetails]': ...

class PydanticValueError(ValueError):
    kind: str
    message_template: str
    context: 'dict[str, str | int] | None'

    def __init__(self, kind: str, message_template: str, context: 'dict[str, str | int] | None' = None) -> None: ...
    def message(self) -> str: ...



import pytest

from pydantic_core import SchemaError, SchemaValidator, ValidationError


@pytest.mark.parametrize(
    'input_value,expected_value',
    [
        (True, True),
        (False, False),
        ('true', True),
        ('false', False),
        (1, 1),
        (0, 0),
        (123, 123),
        ('123', 123),
        ('0', False),  # this case is different depending on the order of the choices
        ('1', True),  # this case is different depending on the order of the choices
    ],
)
def test_union_bool_int(input_value, expected_value):
    v = SchemaValidator({'type': 'union', 'choices': [{'type': 'bool'}, {'type': 'int'}]})
    assert v.validate_python(input_value) == expected_value


@pytest.mark.parametrize(
    'input_value,expected_value',
    [
        (True, True),
        (False, False),
        ('true', True),
        ('false', False),
        (1, 1),
        (0, 0),
        (123, 123),
        ('123', 123),
        ('0', 0),  # this case is different depending on the order of the choices
        ('1', 1),  # this case is different depending on the order of the choices
    ],
)
def test_union_int_bool(input_value, expected_value):
    v = SchemaValidator({'type': 'union', 'choices': [{'type': 'int'}, {'type': 'bool'}]})
    assert v.validate_python(input_value) == expected_value


class TestModelClass:
    class ModelA:
        pass

    class ModelB:
        pass

    @pytest.fixture(scope='class')
    def schema_validator(self) -> SchemaValidator:
        return SchemaValidator(
            {
                'type': 'union',
                'choices': [
                    {
                        'type': 'new-class',
                        'class_type': self.ModelA,
                        'schema': {
                            'type': 'typed-dict',
                            'return_fields_set': True,
                            'fields': {'a': {'schema': {'type': 'int'}}, 'b': {'schema': {'type': 'str'}}},
                        },
                    },
                    {
                        'type': 'new-class',
                        'class_type': self.ModelB,
                        'schema': {
                            'type': 'typed-dict',
                            'return_fields_set': True,
                            'fields': {'c': {'schema': {'type': 'int'}}, 'd': {'schema': {'type': 'str'}}},
                        },
                    },
                ],
            }
        )

    def test_model_a(self, schema_validator: SchemaValidator):
        m_a = schema_validator.validate_python({'a': 1, 'b': 'hello'})
        assert isinstance(m_a, self.ModelA)
        assert m_a.a == 1
        assert m_a.b == 'hello'

    def test_model_b(self, schema_validator: SchemaValidator):
        m_b = schema_validator.validate_python({'c': 2, 'd': 'again'})
        assert isinstance(m_b, self.ModelB)
        assert m_b.c == 2
        assert m_b.d == 'again'

    def test_exact_check(self, schema_validator: SchemaValidator):
        m_b = schema_validator.validate_python({'c': 2, 'd': 'again'})
        assert isinstance(m_b, self.ModelB)

        m_b2 = schema_validator.validate_python(m_b)
        assert m_b2 is m_b

    def test_error(self, schema_validator: SchemaValidator):
        with pytest.raises(ValidationError) as exc_info:
            schema_validator.validate_python({'a': 2})
        assert exc_info.value.errors() == [
            {'kind': 'missing', 'loc': ['ModelA', 'b'], 'message': 'Field required', 'input_value': {'a': 2}},
            {'kind': 'missing', 'loc': ['ModelB', 'c'], 'message': 'Field required', 'input_value': {'a': 2}},
            {'kind': 'missing', 'loc': ['ModelB', 'd'], 'message': 'Field required', 'input_value': {'a': 2}},
        ]


class TestModelClassSimilar:
    class ModelA:
        pass

    class ModelB:
        pass

    @pytest.fixture(scope='class')
    def schema_validator(self) -> SchemaValidator:
        return SchemaValidator(
            {
                'type': 'union',
                'choices': [
                    {
                        'type': 'new-class',
                        'class_type': self.ModelA,
                        'schema': {
                            'type': 'typed-dict',
                            'return_fields_set': True,
                            'fields': {'a': {'schema': {'type': 'int'}}, 'b': {'schema': {'type': 'str'}}},
                        },
                    },
                    {
                        'type': 'new-class',
                        'class_type': self.ModelB,
                        'schema': {
                            'type': 'typed-dict',
                            'return_fields_set': True,
                            'fields': {
                                'a': {'schema': {'type': 'int'}},
                                'b': {'schema': {'type': 'str'}},
                                'c': {'schema': {'type': 'float'}, 'default': 1.0},
                            },
                        },
                    },
                ],
            }
        )

    def test_model_a(self, schema_validator: SchemaValidator):
        m = schema_validator.validate_python({'a': 1, 'b': 'hello'})
        assert isinstance(m, self.ModelA)
        assert m.a == 1
        assert m.b == 'hello'
        assert not hasattr(m, 'c')

    def test_model_b_ignored(self, schema_validator: SchemaValidator):
        # first choice works, so second choice is not used
        m = schema_validator.validate_python({'a': 1, 'b': 'hello', 'c': 2.0})
        assert isinstance(m, self.ModelA)
        assert m.a == 1
        assert m.b == 'hello'
        assert not hasattr(m, 'c')

    def test_model_b_not_ignored(self, schema_validator: SchemaValidator):
        m1 = self.ModelB()
        m1.a = 1
        m1.b = 'hello'
        m1.c = 2.0
        m2 = schema_validator.validate_python(m1)
        assert isinstance(m2, self.ModelB)
        assert m2.a == 1
        assert m2.b == 'hello'
        assert m2.c == 2.0


def test_nullable_via_union():
    v = SchemaValidator({'type': 'union', 'choices': [{'type': 'none'}, {'type': 'int'}]})
    assert v.validate_python(None) is None
    assert v.validate_python(1) == 1
    with pytest.raises(ValidationError) as exc_info:
        v.validate_python('hello')
    assert exc_info.value.errors() == [
        {'kind': 'none_required', 'loc': ['none'], 'message': 'Input should be None/null', 'input_value': 'hello'},
        {
            'kind': 'int_parsing',
            'loc': ['int'],
            'message': 'Input should be a valid integer, unable to parse string as an integer',
            'input_value': 'hello',
        },
    ]


def test_union_list_bool_int():
    v = SchemaValidator(
        {
            'type': 'union',
            'choices': [
                {'type': 'list', 'items_schema': {'type': 'bool'}},
                {'type': 'list', 'items_schema': {'type': 'int'}},
            ],
        }
    )
    assert v.validate_python(['true', True, 'no']) == [True, True, False]
    assert v.validate_python([5, 6, '789']) == [5, 6, 789]
    assert v.validate_python(['1', '0']) == [1, 0]
    with pytest.raises(ValidationError) as exc_info:
        v.validate_python([3, 'true'])
    assert exc_info.value.errors() == [
        {
            'kind': 'bool_parsing',
            'loc': ['list[bool]', 0],
            'message': 'Input should be a valid boolean, unable to interpret input',
            'input_value': 3,
        },
        {
            'kind': 'int_parsing',
            'loc': ['list[int]', 1],
            'message': 'Input should be a valid integer, unable to parse string as an integer',
            'input_value': 'true',
        },
    ]


def test_no_choices():
    with pytest.raises(SchemaError) as exc_info:
        SchemaValidator({'type': 'union'})

    assert exc_info.value.args[0] == (
        'Invalid Schema:\n'
        'union -> choices\n'
        "  Field required [kind=missing, input_value={'type': 'union'}, input_type=dict]"
    )


def test_strict_union():
    v = SchemaValidator({'type': 'union', 'strict': True, 'choices': [{'type': 'bool'}, {'type': 'int'}]})
    assert v.validate_python(1) == 1
    assert v.validate_python(123) == 123

    with pytest.raises(ValidationError) as exc_info:
        v.validate_python('123')

    assert exc_info.value.errors() == [
        {'kind': 'bool_type', 'loc': ['bool'], 'message': 'Input should be a valid boolean', 'input_value': '123'},
        {'kind': 'int_type', 'loc': ['int'], 'message': 'Input should be a valid integer', 'input_value': '123'},
    ]



import re
import sys
from pathlib import Path

import pytest

from pydantic_core._pydantic_core import SchemaError, SchemaValidator, ValidationError, __version__


@pytest.mark.parametrize('obj', [ValidationError, SchemaValidator, SchemaError])
def test_module(obj):
    assert obj.__module__ == 'pydantic_core._pydantic_core'


def test_version():
    assert isinstance(__version__, str)
    assert '.' in __version__


def test_schema_error():
    err = SchemaError('test')
    assert isinstance(err, Exception)
    assert str(err) == 'test'
    assert repr(err) == 'SchemaError("test")'


def test_validation_error():
    v = SchemaValidator('int')
    with pytest.raises(ValidationError) as exc_info:
        v.validate_python(1.5)

    assert exc_info.value.title == 'int'
    assert exc_info.value.error_count() == 1


def test_validation_error_multiple():
    class MyModel:
        # this is not required, but it avoids `__fields_set__` being included in `__dict__`
        __slots__ = '__dict__', '__fields_set__'
        field_a: str
        field_b: int

    v = SchemaValidator(
        {
            'type': 'new-class',
            'class_type': MyModel,
            'schema': {
                'type': 'typed-dict',
                'return_fields_set': True,
                'fields': {'x': {'schema': {'type': 'float'}}, 'y': {'schema': {'type': 'int'}}},
            },
        }
    )
    with pytest.raises(ValidationError) as exc_info:
        v.validate_python({'x': 'x' * 60, 'y': 'y'})

    assert exc_info.value.title == 'MyModel'
    assert exc_info.value.error_count() == 2
    assert exc_info.value.errors() == [
        {
            'kind': 'float_parsing',
            'loc': ['x'],
            'message': 'Input should be a valid number, unable to parse string as an number',
            'input_value': 'x' * 60,
        },
        {
            'kind': 'int_parsing',
            'loc': ['y'],
            'message': 'Input should be a valid integer, unable to parse string as an integer',
            'input_value': 'y',
        },
    ]
    assert repr(exc_info.value) == (
        '2 validation errors for MyModel\n'
        'x\n'
        '  Input should be a valid number, unable to parse string as an number [kind=float_parsing, '
        "input_value='xxxxxxxxxxxxxxxxxxxxxxxx...xxxxxxxxxxxxxxxxxxxxxxx', input_type=str]\n"
        'y\n'
        "  Input should be a valid integer, unable to parse string as an integer [kind=int_parsing, input_value='y', "
        'input_type=str]'
    )


@pytest.mark.skipif(sys.platform == 'emscripten', reason='README.md is not mounted in wasm file system')
def test_readme(import_execute):
    this_dir = Path(__file__).parent
    readme = (this_dir / '..' / 'README.md').read_text()
    example_code = re.search(r'\n```py\n(.*?)\n```\n', readme, re.M | re.S).group(1)
    import_execute(example_code)



import re
from collections import OrderedDict
from collections.abc import Mapping
from typing import Any, Dict

import pytest
from dirty_equals import HasRepr, IsStr

from pydantic_core import SchemaValidator, ValidationError

from ..conftest import Err, PyAndJson


def test_dict(py_and_json: PyAndJson):
    v = py_and_json({'type': 'dict', 'keys_schema': {'type': 'int'}, 'values_schema': {'type': 'int'}})
    assert v.validate_test({'1': 2, '3': 4}) == {1: 2, 3: 4}
    v = py_and_json({'type': 'dict', 'strict': True, 'keys_schema': {'type': 'int'}, 'values_schema': {'type': 'int'}})
    assert v.validate_test({'1': 2, '3': 4}) == {1: 2, 3: 4}
    assert v.validate_test({}) == {}
    with pytest.raises(ValidationError, match='Input should be a valid dictionary'):
        v.validate_test([])


@pytest.mark.parametrize(
    'input_value,expected',
    [
        ({'1': b'1', '2': b'2'}, {'1': '1', '2': '2'}),
        (OrderedDict(a=b'1', b='2'), {'a': '1', 'b': '2'}),
        ({}, {}),
        ('foobar', Err("Input should be a valid dictionary [kind=dict_type, input_value='foobar', input_type=str]")),
        ([], Err('Input should be a valid dictionary [kind=dict_type,')),
        ([('x', 'y')], Err('Input should be a valid dictionary [kind=dict_type,')),
        ([('x', 'y'), ('z', 'z')], Err('Input should be a valid dictionary [kind=dict_type,')),
        ((), Err('Input should be a valid dictionary [kind=dict_type,')),
        ((('x', 'y'),), Err('Input should be a valid dictionary [kind=dict_type,')),
        ((type('Foobar', (), {'x': 1})()), Err('Input should be a valid dictionary [kind=dict_type,')),
    ],
    ids=repr,
)
def test_dict_cases(input_value, expected):
    v = SchemaValidator({'type': 'dict', 'keys_schema': 'str', 'values_schema': 'str'})
    if isinstance(expected, Err):
        with pytest.raises(ValidationError, match=re.escape(expected.message)):
            v.validate_python(input_value)
    else:
        assert v.validate_python(input_value) == expected


def test_dict_value_error(py_and_json: PyAndJson):
    v = py_and_json({'type': 'dict', 'values_schema': 'int'})
    assert v.validate_test({'a': 2, 'b': '4'}) == {'a': 2, 'b': 4}
    with pytest.raises(ValidationError, match='Input should be a valid integer') as exc_info:
        v.validate_test({'a': 2, 'b': 'wrong'})
    assert exc_info.value.errors() == [
        {
            'kind': 'int_parsing',
            'loc': ['b'],
            'message': 'Input should be a valid integer, unable to parse string as an integer',
            'input_value': 'wrong',
        }
    ]


def test_dict_error_key_int():
    v = SchemaValidator({'type': 'dict', 'values_schema': 'int'})
    with pytest.raises(ValidationError, match='Input should be a valid integer') as exc_info:
        v.validate_python({1: 2, 3: 'wrong'})
    assert exc_info.value.errors() == [
        {
            'kind': 'int_parsing',
            'loc': [3],
            'message': 'Input should be a valid integer, unable to parse string as an integer',
            'input_value': 'wrong',
        }
    ]


def test_dict_error_key_other():
    v = SchemaValidator({'type': 'dict', 'values_schema': 'int'})
    with pytest.raises(ValidationError, match='Input should be a valid integer') as exc_info:
        v.validate_python({1: 2, (1, 2): 'wrong'})
    assert exc_info.value.errors() == [
        {
            'kind': 'int_parsing',
            'loc': ['(1, 2)'],
            'message': 'Input should be a valid integer, unable to parse string as an integer',
            'input_value': 'wrong',
        }
    ]


def test_dict_any_value():
    v = SchemaValidator({'type': 'dict', 'keys_schema': {'type': 'str'}})
    v = SchemaValidator({'type': 'dict', 'keys_schema': {'type': 'str'}})
    assert v.validate_python({'1': 1, '2': 'a', '3': None}) == {'1': 1, '2': 'a', '3': None}


def test_mapping():
    class MyMapping(Mapping):
        def __init__(self, d):
            self._d = d

        def __getitem__(self, key):
            return self._d[key]

        def __iter__(self):
            return iter(self._d)

        def __len__(self):
            return len(self._d)

    v = SchemaValidator({'type': 'dict', 'keys_schema': {'type': 'int'}, 'values_schema': {'type': 'int'}})
    assert v.validate_python(MyMapping({'1': 2, 3: '4'})) == {1: 2, 3: 4}
    v = SchemaValidator(
        {'type': 'dict', 'strict': True, 'keys_schema': {'type': 'int'}, 'values_schema': {'type': 'int'}}
    )
    with pytest.raises(ValidationError, match='Input should be a valid dictionary'):
        v.validate_python(MyMapping({'1': 2, 3: '4'}))


def test_key_error():
    v = SchemaValidator({'type': 'dict', 'keys_schema': {'type': 'int'}, 'values_schema': {'type': 'int'}})
    assert v.validate_python({'1': True}) == {1: 1}
    with pytest.raises(ValidationError, match=re.escape('x -> [key]\n  Input should be a valid integer')) as exc_info:
        v.validate_python({'x': 1})
    assert exc_info.value.errors() == [
        {
            'kind': 'int_parsing',
            'loc': ['x', '[key]'],
            'message': 'Input should be a valid integer, unable to parse string as an integer',
            'input_value': 'x',
        }
    ]


def test_mapping_error():
    class BadMapping(Mapping):
        def __getitem__(self, key):
            raise None

        def __iter__(self):
            raise RuntimeError('intentional error')

        def __len__(self):
            return 1

    v = SchemaValidator({'type': 'dict', 'keys_schema': {'type': 'int'}, 'values_schema': {'type': 'int'}})
    with pytest.raises(ValidationError) as exc_info:
        v.validate_python(BadMapping())

    assert exc_info.value.errors() == [
        {
            'kind': 'dict_from_mapping',
            'loc': [],
            'message': 'Unable to convert mapping to a dictionary, error: RuntimeError: intentional error',
            'input_value': HasRepr(IsStr(regex='.+BadMapping object at.+')),
            'context': {'error': 'RuntimeError: intentional error'},
        }
    ]


def test_mapping_error_yield_1():
    class BadMapping(Mapping):
        def items(self):
            return [(1,)]

        def __iter__(self):
            return iter({1: 2})

        def __getitem__(self, key):
            raise None

        def __len__(self):
            return 1

    v = SchemaValidator({'type': 'dict', 'keys_schema': {'type': 'int'}, 'values_schema': {'type': 'int'}})
    with pytest.raises(ValidationError) as exc_info:
        v.validate_python(BadMapping())

    assert exc_info.value.errors() == [
        {
            'kind': 'dict_from_mapping',
            'loc': [],
            'message': (
                'Unable to convert mapping to a dictionary, error: '
                'ValueError: Expected tuple of length 2, but got tuple of length 1.'
            ),
            'input_value': HasRepr(IsStr(regex='.+BadMapping object at.+')),
            'context': {'error': 'ValueError: Expected tuple of length 2, but got tuple of length 1.'},
        }
    ]


@pytest.mark.parametrize(
    'kwargs,input_value,expected',
    [
        ({}, {'1': 1, '2': 2}, {'1': 1, '2': 2}),
        (
            {'min_items': 3},
            {'1': 1, '2': 2, '3': 3.0, '4': [1, 2, 3, 4]},
            {'1': 1, '2': 2, '3': 3.0, '4': [1, 2, 3, 4]},
        ),
        ({'min_items': 3}, {1: '2', 3: '4'}, Err('Input should have at least 3 items, got 2 items [kind=too_short,')),
        ({'max_items': 4}, {'1': 1, '2': 2, '3': 3.0}, {'1': 1, '2': 2, '3': 3.0}),
        (
            {'max_items': 3},
            {'1': 1, '2': 2, '3': 3.0, '4': [1, 2, 3, 4]},
            Err('Input should have at most 3 items, got 4 items [kind=too_long,'),
        ),
    ],
)
def test_dict_length_constraints(kwargs: Dict[str, Any], input_value, expected):
    v = SchemaValidator({'type': 'dict', **kwargs})
    if isinstance(expected, Err):
        with pytest.raises(ValidationError, match=re.escape(expected.message)):
            v.validate_python(input_value)
    else:
        assert v.validate_python(input_value) == expected



from ._pydantic_core import PydanticValueError, SchemaError, SchemaValidator, ValidationError, __version__
from ._types import Config, Schema

__all__ = '__version__', 'Config', 'Schema', 'SchemaValidator', 'SchemaError', 'ValidationError', 'PydanticValueError'



import os
import sys
from pathlib import Path


def error(msg: str):
    print(f'ERROR: {msg}', file=sys.stderr)
    exit(1)


try:
    import requests
except ImportError:
    error("requests not installed, you'll need to run `pip install requests`")


def main():
    root_dir = Path(__file__).parent.parent
    try:
        wheel_file = next(p for p in (root_dir / 'dist').iterdir() if p.name.endswith('wasm32.whl'))
    except StopIteration:
        error('No wheel found in "dist" directory')
    else:
        uploader = Uploader()

        wheel_url = uploader.upload_file(wheel_file)
        print(f'Wheel uploaded ✓, URL: "{wheel_url}"')


class Uploader:
    def __init__(self):
        try:
            auth_key = os.environ['SMOKESHOW_AUTH_KEY']
        except KeyError:
            raise RuntimeError('No auth key provided, please set SMOKESHOW_AUTH_KEY')
        else:
            self.client = requests.Session()
            r = self.client.post('https://smokeshow.helpmanual.io/create/', headers={'Authorisation': auth_key})
            if r.status_code != 200:
                raise ValueError(f'Error creating ephemeral site {r.status_code}, response:\n{r.text}')

            obj = r.json()
            self.secret_key: str = obj['secret_key']
            self.url: str = obj['url']
            assert self.url.endswith('/'), self.url

    def upload_file(self, file: Path) -> str:
        headers = {'Authorisation': self.secret_key, 'Response-Header-Access-Control-Allow-Origin': '*'}

        url_path = file.name
        url = self.url + file.name
        r = self.client.post(url, data=file.read_bytes(), headers=headers)
        if r.status_code == 200:
            upload_info = r.json()
            print(f'    uploaded {url_path} size={upload_info["size"]:,}')
        else:
            print(f'    ERROR! {url_path} status={r.status_code} response={r.text}')
            error(f'invalid response from "{url_path}" status={r.status_code} response={r.text}')
        return url


if __name__ == '__main__':
    main()



import base64
import re
import sys
import importlib
import traceback
from io import BytesIO
from pathlib import Path
from zipfile import ZipFile

import micropip
import pytest

# this seems to be required for me on M1 Mac
sys.setrecursionlimit(200)

# compiled manually an uploaded to smokeshow, there seems to be no nice way of getting a file from a CI build
pydantic_core_wheel = (
    'https://smokeshow.helpmanual.io'
    '/4o4l4x0t2m6z1w4n6u4b/pydantic_core-0.0.1-cp310-cp310-emscripten_3_1_14_wasm32.whl'
)


async def main(tests_zip: str):
    print(f'Extracting test files (size: {len(tests_zip):,})...')
    zip_file = ZipFile(BytesIO(base64.b64decode(tests_zip)))
    count = 0
    for name in zip_file.namelist():
        if name.endswith('.py'):
            path, subs = re.subn(r'^pydantic-core-.+?/tests/', 'tests/', name)
            if subs:
                count += 1
                path = Path(path)
                path.parent.mkdir(parents=True, exist_ok=True)
                with zip_file.open(name, 'r') as f:
                    path.write_bytes(f.read())

    print(f'Mounted {count} test files, installing dependencies...')

    await micropip.install(['dirty-equals', 'hypothesis', 'pytest-speed', pydantic_core_wheel])
    importlib.invalidate_caches()

    # print('installed packages:')
    # print(micropip.list())
    print('Running tests...')
    pytest.main()

try:
    await main(tests_zip)
except Exception as e:
    traceback.print_exc()
    raise



# this file is auto-generated by generate_self_schema.py, DO NOT edit manually
self_schema = {
    'type': 'tagged-union',
    'ref': 'root-schema',
    'discriminator': 'self-schema-discriminator',
    'choices': {
        'plain-string': {
            'type': 'literal',
            'expected': [
                'any',
                'none',
                'str',
                'bytes',
                'dict',
                'int',
                'bool',
                'float',
                'list',
                'tuple',
                'set',
                'frozenset',
                'date',
                'time',
                'datetime',
                'timedelta',
                'callable',
            ],
        },
        'any': {
            'type': 'typed-dict',
            'description': 'AnySchema',
            'fields': {'type': {'schema': {'type': 'literal', 'expected': ['any']}, 'required': True}},
            'extra_behavior': 'forbid',
        },
        'bool': {
            'type': 'typed-dict',
            'description': 'BoolSchema',
            'fields': {
                'type': {'schema': {'type': 'literal', 'expected': ['bool']}, 'required': True},
                'strict': {'schema': 'bool', 'required': False},
                'ref': {'schema': 'str', 'required': False},
            },
            'extra_behavior': 'forbid',
        },
        'bytes': {
            'type': 'typed-dict',
            'description': 'BytesSchema',
            'fields': {
                'type': {'schema': {'type': 'literal', 'expected': ['bytes']}, 'required': True},
                'max_length': {'schema': 'int', 'required': False},
                'min_length': {'schema': 'int', 'required': False},
                'strict': {'schema': 'bool', 'required': False},
                'ref': {'schema': 'str', 'required': False},
            },
            'extra_behavior': 'forbid',
        },
        'dict': {
            'type': 'typed-dict',
            'description': 'DictSchema',
            'fields': {
                'type': {'schema': {'type': 'literal', 'expected': ['dict']}, 'required': True},
                'keys_schema': {'schema': {'type': 'recursive-ref', 'schema_ref': 'root-schema'}, 'required': False},
                'values_schema': {'schema': {'type': 'recursive-ref', 'schema_ref': 'root-schema'}, 'required': False},
                'min_items': {'schema': 'int', 'required': False},
                'max_items': {'schema': 'int', 'required': False},
                'strict': {'schema': 'bool', 'required': False},
                'ref': {'schema': 'str', 'required': False},
            },
            'extra_behavior': 'forbid',
        },
        'float': {
            'type': 'typed-dict',
            'description': 'FloatSchema',
            'fields': {
                'type': {'schema': {'type': 'literal', 'expected': ['float']}, 'required': True},
                'multiple_of': {'schema': 'float', 'required': False},
                'le': {'schema': 'float', 'required': False},
                'ge': {'schema': 'float', 'required': False},
                'lt': {'schema': 'float', 'required': False},
                'gt': {'schema': 'float', 'required': False},
                'strict': {'schema': 'bool', 'required': False},
                'ref': {'schema': 'str', 'required': False},
            },
            'extra_behavior': 'forbid',
        },
        'function': {
            'type': 'typed-dict',
            'description': 'FunctionSchema',
            'fields': {
                'type': {'schema': {'type': 'literal', 'expected': ['function']}, 'required': True},
                'mode': {'schema': {'type': 'literal', 'expected': ['before', 'after', 'wrap']}, 'required': True},
                'function': {'schema': 'callable', 'required': True},
                'schema': {'schema': {'type': 'recursive-ref', 'schema_ref': 'root-schema'}, 'required': False},
                'ref': {'schema': 'str', 'required': False},
            },
            'extra_behavior': 'forbid',
        },
        'function-plain': {
            'type': 'typed-dict',
            'description': 'FunctionPlainSchema',
            'fields': {
                'type': {'schema': {'type': 'literal', 'expected': ['function']}, 'required': True},
                'mode': {'schema': {'type': 'literal', 'expected': ['plain']}, 'required': True},
                'function': {'schema': 'callable', 'required': True},
                'ref': {'schema': 'str', 'required': False},
            },
            'extra_behavior': 'forbid',
        },
        'int': {
            'type': 'typed-dict',
            'description': 'IntSchema',
            'fields': {
                'type': {'schema': {'type': 'literal', 'expected': ['int']}, 'required': True},
                'multiple_of': {'schema': 'int', 'required': False},
                'le': {'schema': 'int', 'required': False},
                'ge': {'schema': 'int', 'required': False},
                'lt': {'schema': 'int', 'required': False},
                'gt': {'schema': 'int', 'required': False},
                'strict': {'schema': 'bool', 'required': False},
                'ref': {'schema': 'str', 'required': False},
            },
            'extra_behavior': 'forbid',
        },
        'list': {
            'type': 'typed-dict',
            'description': 'ListSchema',
            'fields': {
                'type': {'schema': {'type': 'literal', 'expected': ['list']}, 'required': True},
                'items_schema': {'schema': {'type': 'recursive-ref', 'schema_ref': 'root-schema'}, 'required': False},
                'min_items': {'schema': 'int', 'required': False},
                'max_items': {'schema': 'int', 'required': False},
                'strict': {'schema': 'bool', 'required': False},
                'ref': {'schema': 'str', 'required': False},
            },
            'extra_behavior': 'forbid',
        },
        'literal': {
            'type': 'typed-dict',
            'description': 'LiteralSchema',
            'fields': {
                'type': {'schema': {'type': 'literal', 'expected': ['literal']}, 'required': True},
                'expected': {'schema': {'type': 'list', 'items_schema': 'any'}, 'required': True},
                'ref': {'schema': 'str', 'required': False},
            },
            'extra_behavior': 'forbid',
        },
        'typed-dict': {
            'type': 'typed-dict',
            'description': 'TypedDictSchema',
            'fields': {
                'type': {'schema': {'type': 'literal', 'expected': ['typed-dict']}, 'required': True},
                'fields': {
                    'schema': {
                        'type': 'dict',
                        'keys_schema': 'str',
                        'values_schema': {
                            'type': 'typed-dict',
                            'description': 'TypedDictField',
                            'fields': {
                                'schema': {
                                    'schema': {'type': 'recursive-ref', 'schema_ref': 'root-schema'},
                                    'required': True,
                                },
                                'required': {'schema': 'bool', 'required': False},
                                'default': {'schema': 'any', 'required': False},
                                'default_factory': {'schema': 'callable', 'required': False},
                                'on_error': {
                                    'schema': {'type': 'literal', 'expected': ['raise', 'omit', 'fallback_on_default']},
                                    'required': False,
                                },
                                'alias': {
                                    'schema': {
                                        'type': 'union',
                                        'choices': [
                                            'str',
                                            {
                                                'type': 'list',
                                                'items_schema': {'type': 'union', 'choices': ['str', 'int']},
                                            },
                                            {
                                                'type': 'list',
                                                'items_schema': {
                                                    'type': 'list',
                                                    'items_schema': {'type': 'union', 'choices': ['str', 'int']},
                                                },
                                            },
                                        ],
                                    },
                                    'required': False,
                                },
                            },
                            'extra_behavior': 'forbid',
                        },
                    },
                    'required': True,
                },
                'strict': {'schema': 'bool', 'required': False},
                'extra_validator': {
                    'schema': {'type': 'recursive-ref', 'schema_ref': 'root-schema'},
                    'required': False,
                },
                'return_fields_set': {'schema': 'bool', 'required': False},
                'ref': {'schema': 'str', 'required': False},
                'extra_behavior': {
                    'schema': {'type': 'literal', 'expected': ['allow', 'forbid', 'ignore']},
                    'required': False,
                },
                'full': {'schema': 'bool', 'required': False},
                'populate_by_name': {'schema': 'bool', 'required': False},
                'from_attributes': {'schema': 'bool', 'required': False},
            },
            'extra_behavior': 'forbid',
        },
        'new-class': {
            'type': 'typed-dict',
            'description': 'NewClassSchema',
            'fields': {
                'type': {'schema': {'type': 'literal', 'expected': ['new-class']}, 'required': True},
                'class_type': {'schema': 'any', 'required': True},
                'schema': {'schema': {'type': 'recursive-ref', 'schema_ref': 'root-schema'}, 'required': True},
                'strict': {'schema': 'bool', 'required': False},
                'ref': {'schema': 'str', 'required': False},
                'config': {
                    'schema': {
                        'type': 'typed-dict',
                        'description': 'Config',
                        'fields': {
                            'strict': {'schema': 'bool', 'required': False},
                            'config_choose_priority': {'schema': 'int', 'required': False},
                            'config_merge_priority': {'schema': 'int', 'required': False},
                            'typed_dict_extra_behavior': {
                                'schema': {'type': 'literal', 'expected': ['allow', 'forbid', 'ignore']},
                                'required': False,
                            },
                            'typed_dict_full': {'schema': 'bool', 'required': False},
                            'from_attributes': {'schema': 'bool', 'required': False},
                            'revalidate_models': {'schema': 'bool', 'required': False},
                            'populate_by_name': {'schema': 'bool', 'required': False},
                            'str_max_length': {'schema': 'int', 'required': False},
                            'str_min_length': {'schema': 'int', 'required': False},
                            'str_strip_whitespace': {'schema': 'bool', 'required': False},
                            'str_to_lower': {'schema': 'bool', 'required': False},
                            'str_to_upper': {'schema': 'bool', 'required': False},
                        },
                        'extra_behavior': 'forbid',
                    },
                    'required': False,
                },
            },
            'extra_behavior': 'forbid',
        },
        'none': {
            'type': 'typed-dict',
            'description': 'NoneSchema',
            'fields': {
                'type': {'schema': {'type': 'literal', 'expected': ['none']}, 'required': True},
                'ref': {'schema': 'str', 'required': False},
            },
            'extra_behavior': 'forbid',
        },
        'nullable': {
            'type': 'typed-dict',
            'description': 'NullableSchema',
            'fields': {
                'type': {'schema': {'type': 'literal', 'expected': ['nullable']}, 'required': True},
                'schema': {'schema': {'type': 'recursive-ref', 'schema_ref': 'root-schema'}, 'required': True},
                'strict': {'schema': 'bool', 'required': False},
                'ref': {'schema': 'str', 'required': False},
            },
            'extra_behavior': 'forbid',
        },
        'recursive-ref': {
            'type': 'typed-dict',
            'description': 'RecursiveReferenceSchema',
            'fields': {
                'type': {'schema': {'type': 'literal', 'expected': ['recursive-ref']}, 'required': True},
                'schema_ref': {'schema': 'str', 'required': True},
            },
            'extra_behavior': 'forbid',
        },
        'set': {
            'type': 'typed-dict',
            'description': 'SetSchema',
            'fields': {
                'type': {'schema': {'type': 'literal', 'expected': ['set']}, 'required': True},
                'items_schema': {'schema': {'type': 'recursive-ref', 'schema_ref': 'root-schema'}, 'required': False},
                'min_items': {'schema': 'int', 'required': False},
                'max_items': {'schema': 'int', 'required': False},
                'strict': {'schema': 'bool', 'required': False},
                'ref': {'schema': 'str', 'required': False},
            },
            'extra_behavior': 'forbid',
        },
        'frozenset': {
            'type': 'typed-dict',
            'description': 'FrozenSetSchema',
            'fields': {
                'type': {'schema': {'type': 'literal', 'expected': ['frozenset']}, 'required': True},
                'items_schema': {'schema': {'type': 'recursive-ref', 'schema_ref': 'root-schema'}, 'required': False},
                'min_items': {'schema': 'int', 'required': False},
                'max_items': {'schema': 'int', 'required': False},
                'strict': {'schema': 'bool', 'required': False},
                'ref': {'schema': 'str', 'required': False},
            },
            'extra_behavior': 'forbid',
        },
        'str': {
            'type': 'typed-dict',
            'description': 'StringSchema',
            'fields': {
                'type': {'schema': {'type': 'literal', 'expected': ['str']}, 'required': True},
                'pattern': {'schema': 'str', 'required': False},
                'max_length': {'schema': 'int', 'required': False},
                'min_length': {'schema': 'int', 'required': False},
                'strip_whitespace': {'schema': 'bool', 'required': False},
                'to_lower': {'schema': 'bool', 'required': False},
                'to_upper': {'schema': 'bool', 'required': False},
                'strict': {'schema': 'bool', 'required': False},
                'ref': {'schema': 'str', 'required': False},
            },
            'extra_behavior': 'forbid',
        },
        'tuple-positional': {
            'type': 'typed-dict',
            'description': 'TuplePositionalSchema',
            'fields': {
                'type': {'schema': {'type': 'literal', 'expected': ['tuple']}, 'required': True},
                'mode': {'schema': {'type': 'literal', 'expected': ['positional']}, 'required': True},
                'items_schema': {
                    'schema': {'type': 'list', 'items_schema': {'type': 'recursive-ref', 'schema_ref': 'root-schema'}},
                    'required': True,
                },
                'extra_schema': {'schema': {'type': 'recursive-ref', 'schema_ref': 'root-schema'}, 'required': False},
                'strict': {'schema': 'bool', 'required': False},
                'ref': {'schema': 'str', 'required': False},
            },
            'extra_behavior': 'forbid',
        },
        'tuple-variable': {
            'type': 'typed-dict',
            'description': 'TupleVariableSchema',
            'fields': {
                'type': {'schema': {'type': 'literal', 'expected': ['tuple']}, 'required': True},
                'mode': {'schema': {'type': 'literal', 'expected': ['variable']}, 'required': False},
                'items_schema': {'schema': {'type': 'recursive-ref', 'schema_ref': 'root-schema'}, 'required': False},
                'min_items': {'schema': 'int', 'required': False},
                'max_items': {'schema': 'int', 'required': False},
                'strict': {'schema': 'bool', 'required': False},
                'ref': {'schema': 'str', 'required': False},
            },
            'extra_behavior': 'forbid',
        },
        'union': {
            'type': 'typed-dict',
            'description': 'UnionSchema',
            'fields': {
                'type': {'schema': {'type': 'literal', 'expected': ['union']}, 'required': True},
                'choices': {
                    'schema': {'type': 'list', 'items_schema': {'type': 'recursive-ref', 'schema_ref': 'root-schema'}},
                    'required': True,
                },
                'strict': {'schema': 'bool', 'required': False},
                'ref': {'schema': 'str', 'required': False},
            },
            'extra_behavior': 'forbid',
        },
        'tagged-union': {
            'type': 'typed-dict',
            'description': 'TaggedUnionSchema',
            'fields': {
                'type': {'schema': {'type': 'literal', 'expected': ['tagged-union']}, 'required': True},
                'choices': {
                    'schema': {
                        'type': 'dict',
                        'keys_schema': 'str',
                        'values_schema': {'type': 'recursive-ref', 'schema_ref': 'root-schema'},
                    },
                    'required': True,
                },
                'discriminator': {
                    'schema': {
                        'type': 'union',
                        'choices': [
                            'str',
                            {'type': 'list', 'items_schema': {'type': 'union', 'choices': ['str', 'int']}},
                            {
                                'type': 'list',
                                'items_schema': {
                                    'type': 'list',
                                    'items_schema': {'type': 'union', 'choices': ['str', 'int']},
                                },
                            },
                            'callable',
                        ],
                    },
                    'required': True,
                },
                'strict': {'schema': 'bool', 'required': False},
                'ref': {'schema': 'str', 'required': False},
            },
            'extra_behavior': 'forbid',
        },
        'date': {
            'type': 'typed-dict',
            'description': 'DateSchema',
            'fields': {
                'type': {'schema': {'type': 'literal', 'expected': ['date']}, 'required': True},
                'strict': {'schema': 'bool', 'required': False},
                'le': {'schema': 'date', 'required': False},
                'ge': {'schema': 'date', 'required': False},
                'lt': {'schema': 'date', 'required': False},
                'gt': {'schema': 'date', 'required': False},
                'ref': {'schema': 'str', 'required': False},
            },
            'extra_behavior': 'forbid',
        },
        'time': {
            'type': 'typed-dict',
            'description': 'TimeSchema',
            'fields': {
                'type': {'schema': {'type': 'literal', 'expected': ['time']}, 'required': True},
                'strict': {'schema': 'bool', 'required': False},
                'le': {'schema': 'time', 'required': False},
                'ge': {'schema': 'time', 'required': False},
                'lt': {'schema': 'time', 'required': False},
                'gt': {'schema': 'time', 'required': False},
                'ref': {'schema': 'str', 'required': False},
            },
            'extra_behavior': 'forbid',
        },
        'datetime': {
            'type': 'typed-dict',
            'description': 'DatetimeSchema',
            'fields': {
                'type': {'schema': {'type': 'literal', 'expected': ['datetime']}, 'required': True},
                'strict': {'schema': 'bool', 'required': False},
                'le': {'schema': 'datetime', 'required': False},
                'ge': {'schema': 'datetime', 'required': False},
                'lt': {'schema': 'datetime', 'required': False},
                'gt': {'schema': 'datetime', 'required': False},
                'ref': {'schema': 'str', 'required': False},
            },
            'extra_behavior': 'forbid',
        },
        'timedelta': {
            'type': 'typed-dict',
            'description': 'TimedeltaSchema',
            'fields': {
                'type': {'schema': {'type': 'literal', 'expected': ['timedelta']}, 'required': True},
                'strict': {'schema': 'bool', 'required': False},
                'le': {'schema': 'timedelta', 'required': False},
                'ge': {'schema': 'timedelta', 'required': False},
                'lt': {'schema': 'timedelta', 'required': False},
                'gt': {'schema': 'timedelta', 'required': False},
                'ref': {'schema': 'str', 'required': False},
            },
            'extra_behavior': 'forbid',
        },
        'is-instance': {
            'type': 'typed-dict',
            'description': 'IsInstanceSchema',
            'fields': {
                'type': {'schema': {'type': 'literal', 'expected': ['is-instance']}, 'required': True},
                'class_': {'schema': 'any', 'required': True},
            },
            'extra_behavior': 'forbid',
        },
        'callable': {
            'type': 'typed-dict',
            'description': 'CallableSchema',
            'fields': {'type': {'schema': {'type': 'literal', 'expected': ['callable']}, 'required': True}},
            'extra_behavior': 'forbid',
        },
        'arguments': {
            'type': 'typed-dict',
            'description': 'ArgumentsSchema',
            'fields': {
                'type': {'schema': {'type': 'literal', 'expected': ['arguments']}, 'required': True},
                'arguments_schema': {
                    'schema': {
                        'type': 'list',
                        'items_schema': {
                            'type': 'typed-dict',
                            'description': 'Parameter',
                            'fields': {
                                'name': {'schema': 'str', 'required': True},
                                'mode': {
                                    'schema': {
                                        'type': 'literal',
                                        'expected': ['positional_only', 'positional_or_keyword', 'keyword_only'],
                                    },
                                    'required': False,
                                },
                                'schema': {
                                    'schema': {'type': 'recursive-ref', 'schema_ref': 'root-schema'},
                                    'required': True,
                                },
                                'default': {'schema': 'any', 'required': False},
                                'default_factory': {'schema': 'callable', 'required': False},
                                'alias': {
                                    'schema': {
                                        'type': 'union',
                                        'choices': [
                                            'str',
                                            {
                                                'type': 'list',
                                                'items_schema': {'type': 'union', 'choices': ['str', 'int']},
                                            },
                                            {
                                                'type': 'list',
                                                'items_schema': {
                                                    'type': 'list',
                                                    'items_schema': {'type': 'union', 'choices': ['str', 'int']},
                                                },
                                            },
                                        ],
                                    },
                                    'required': False,
                                },
                            },
                            'extra_behavior': 'forbid',
                        },
                    },
                    'required': True,
                },
                'populate_by_name': {'schema': 'bool', 'required': False},
                'var_args_schema': {
                    'schema': {'type': 'recursive-ref', 'schema_ref': 'root-schema'},
                    'required': False,
                },
                'var_kwargs_schema': {
                    'schema': {'type': 'recursive-ref', 'schema_ref': 'root-schema'},
                    'required': False,
                },
                'ref': {'schema': 'str', 'required': False},
            },
            'extra_behavior': 'forbid',
        },
        'call': {
            'type': 'typed-dict',
            'description': 'CallSchema',
            'fields': {
                'type': {'schema': {'type': 'literal', 'expected': ['call']}, 'required': True},
                'function': {'schema': 'callable', 'required': True},
                'arguments_schema': {
                    'schema': {'type': 'recursive-ref', 'schema_ref': 'root-schema'},
                    'required': True,
                },
                'return_schema': {'schema': {'type': 'recursive-ref', 'schema_ref': 'root-schema'}, 'required': False},
                'ref': {'schema': 'str', 'required': False},
            },
            'extra_behavior': 'forbid',
        },
    },
}



from __future__ import annotations

import sys
from datetime import date, datetime, time, timedelta
from typing import Any, Callable, Dict, List, Optional, Type, Union

if sys.version_info < (3, 11):
    from typing_extensions import NotRequired, Required
else:
    from typing import NotRequired, Required

if sys.version_info < (3, 9):
    from typing_extensions import Literal, TypedDict
else:
    from typing import Literal, TypedDict


class AnySchema(TypedDict):
    type: Literal['any']


class BoolSchema(TypedDict, total=False):
    type: Required[Literal['bool']]
    strict: bool
    ref: str


class Config(TypedDict, total=False):
    strict: bool
    # higher priority configs take precedence of over lower, if priority matches the two configs are merged, default 0
    config_choose_priority: int
    # if configs are merged, which should take precedence, default 0, default means child takes precedence
    config_merge_priority: int
    # settings related to typed_dicts only
    typed_dict_extra_behavior: Literal['allow', 'forbid', 'ignore']
    typed_dict_full: bool  # default: True
    # used on typed-dicts and tagged union keys
    from_attributes: bool
    revalidate_models: bool
    # used on typed-dicts and arguments
    populate_by_name: bool  # replaces `allow_population_by_field_name` in pydantic v1
    # fields related to string fields only
    str_max_length: int
    str_min_length: int
    str_strip_whitespace: bool
    str_to_lower: bool
    str_to_upper: bool


class DictSchema(TypedDict, total=False):
    type: Required[Literal['dict']]
    keys_schema: Schema  # default: AnySchema
    values_schema: Schema  # default: AnySchema
    min_items: int
    max_items: int
    strict: bool
    ref: str


class FloatSchema(TypedDict, total=False):
    type: Required[Literal['float']]
    multiple_of: float
    le: float
    ge: float
    lt: float
    gt: float
    strict: bool
    ref: str


class FunctionSchema(TypedDict):
    type: Literal['function']
    mode: Literal['before', 'after', 'wrap']
    function: Callable[..., Any]
    schema: NotRequired[Schema]
    ref: NotRequired[str]


class FunctionPlainSchema(TypedDict):
    type: Literal['function']
    mode: Literal['plain']
    function: Callable[..., Any]
    ref: NotRequired[str]


class IntSchema(TypedDict, total=False):
    type: Required[Literal['int']]
    multiple_of: int
    le: int
    ge: int
    lt: int
    gt: int
    strict: bool
    ref: str


class ListSchema(TypedDict, total=False):
    type: Required[Literal['list']]
    items_schema: Schema  # default: AnySchema
    min_items: int
    max_items: int
    strict: bool
    ref: str


class LiteralSchema(TypedDict):
    type: Literal['literal']
    expected: List[Any]
    ref: NotRequired[str]


class NewClassSchema(TypedDict):
    type: Literal['new-class']
    class_type: type
    schema: Schema
    strict: NotRequired[bool]
    ref: NotRequired[str]
    config: NotRequired[Config]


class TypedDictField(TypedDict, total=False):
    schema: Required[Schema]
    required: bool
    default: Any
    default_factory: Callable[[], Any]
    on_error: Literal['raise', 'omit', 'fallback_on_default']  # default: 'raise'
    alias: Union[str, List[Union[str, int]], List[List[Union[str, int]]]]


class TypedDictSchema(TypedDict, total=False):
    type: Required[Literal['typed-dict']]
    fields: Required[Dict[str, TypedDictField]]
    strict: bool
    extra_validator: Schema
    return_fields_set: bool
    ref: str
    # all these values can be set via config, equivalent fields have `typed_dict_` prefix
    extra_behavior: Literal['allow', 'forbid', 'ignore']
    full: bool  # default: True
    populate_by_name: bool  # replaces `allow_population_by_field_name` in pydantic v1
    from_attributes: bool


class NoneSchema(TypedDict):
    type: Literal['none']
    ref: NotRequired[str]


class NullableSchema(TypedDict, total=False):
    type: Required[Literal['nullable']]
    schema: Required[Schema]
    strict: bool
    ref: str


class RecursiveReferenceSchema(TypedDict):
    type: Literal['recursive-ref']
    schema_ref: str


class SetSchema(TypedDict, total=False):
    type: Required[Literal['set']]
    items_schema: Schema  # default: AnySchema
    min_items: int
    max_items: int
    strict: bool
    ref: str


class FrozenSetSchema(TypedDict, total=False):
    type: Required[Literal['frozenset']]
    items_schema: Schema  # default: AnySchema
    min_items: int
    max_items: int
    strict: bool
    ref: str


class StringSchema(TypedDict, total=False):
    type: Required[Literal['str']]
    pattern: str
    max_length: int
    min_length: int
    strip_whitespace: bool
    to_lower: bool
    to_upper: bool
    strict: bool
    ref: str


class UnionSchema(TypedDict, total=False):
    type: Required[Literal['union']]
    choices: Required[List[Schema]]
    strict: bool
    ref: str


class TaggedUnionSchema(TypedDict):
    type: Literal['tagged-union']
    choices: Dict[str, Schema]
    discriminator: Union[str, List[Union[str, int]], List[List[Union[str, int]]], Callable[[Any], Optional[str]]]
    strict: NotRequired[bool]
    ref: NotRequired[str]


class BytesSchema(TypedDict, total=False):
    type: Required[Literal['bytes']]
    max_length: int
    min_length: int
    strict: bool
    ref: str


class DateSchema(TypedDict, total=False):
    type: Required[Literal['date']]
    strict: bool
    le: date
    ge: date
    lt: date
    gt: date
    ref: str


class TimeSchema(TypedDict, total=False):
    type: Required[Literal['time']]
    strict: bool
    le: time
    ge: time
    lt: time
    gt: time
    ref: str


class DatetimeSchema(TypedDict, total=False):
    type: Required[Literal['datetime']]
    strict: bool
    le: datetime
    ge: datetime
    lt: datetime
    gt: datetime
    ref: str


class TimedeltaSchema(TypedDict, total=False):
    type: Required[Literal['timedelta']]
    strict: bool
    le: timedelta
    ge: timedelta
    lt: timedelta
    gt: timedelta
    ref: str


class TuplePositionalSchema(TypedDict, total=False):
    type: Required[Literal['tuple']]
    mode: Required[Literal['positional']]
    items_schema: Required[List[Schema]]
    extra_schema: Schema
    strict: bool
    ref: str


class TupleVariableSchema(TypedDict, total=False):
    type: Required[Literal['tuple']]
    mode: Literal['variable']
    items_schema: Schema
    min_items: int
    max_items: int
    strict: bool
    ref: str


class IsInstanceSchema(TypedDict):
    type: Literal['is-instance']
    class_: Type[Any]


class CallableSchema(TypedDict):
    type: Literal['callable']


class Parameter(TypedDict, total=False):
    name: Required[str]
    mode: Literal['positional_only', 'positional_or_keyword', 'keyword_only']  # default positional_or_keyword
    schema: Required[Schema]
    default: Any
    default_factory: Callable[[], Any]
    alias: Union[str, List[Union[str, int]], List[List[Union[str, int]]]]


class ArgumentsSchema(TypedDict, total=False):
    type: Required[Literal['arguments']]
    arguments_schema: Required[List[Parameter]]
    populate_by_name: bool
    var_args_schema: Schema
    var_kwargs_schema: Schema
    ref: str


class CallSchema(TypedDict):
    type: Literal['call']
    function: Callable[..., Any]
    arguments_schema: Schema
    return_schema: NotRequired[Schema]
    ref: NotRequired[str]


# pydantic allows types to be defined via a simple string instead of dict with just `type`, e.g.
# 'int' is equivalent to {'type': 'int'}, this only applies to schema types which do not have other required fields
BareType = Literal[
    'any',
    'none',
    'str',
    'bytes',
    'dict',
    'int',
    'bool',
    'float',
    'dict',
    'list',
    'tuple',
    'set',
    'frozenset',
    'date',
    'time',
    'datetime',
    'timedelta',
    'callable',
]

# generate_self_schema.py is hard coded to convert this Union[BareType, Union[...rest]] where the second union is tagged
# so `BareType` MUST come first
Schema = Union[
    BareType,
    AnySchema,
    BoolSchema,
    BytesSchema,
    DictSchema,
    FloatSchema,
    FunctionSchema,
    FunctionPlainSchema,
    IntSchema,
    ListSchema,
    LiteralSchema,
    TypedDictSchema,
    NewClassSchema,
    NoneSchema,
    NullableSchema,
    RecursiveReferenceSchema,
    SetSchema,
    FrozenSetSchema,
    StringSchema,
    TuplePositionalSchema,
    TupleVariableSchema,
    UnionSchema,
    TaggedUnionSchema,
    DateSchema,
    TimeSchema,
    DatetimeSchema,
    TimedeltaSchema,
    IsInstanceSchema,
    CallableSchema,
    ArgumentsSchema,
    CallSchema,
]



import re

import pytest

from pydantic_core import ValidationError

from .conftest import Err, PyAndJson


@pytest.mark.parametrize(
    'strict,input_value,expected',
    [
        (False, 123, 123),
        (False, '123', 123),
        (None, 123, 123),
        (None, '123', 123),
        (True, 123, 123),
        (True, '123', Err('Input should be a valid integer [kind=int_type')),
    ],
)
def test_int_strict_argument(py_and_json: PyAndJson, strict, input_value, expected):
    v = py_and_json({'type': 'int'})
    if isinstance(expected, Err):
        assert v.isinstance_test(input_value, strict) is False
        with pytest.raises(ValidationError, match=re.escape(expected.message)):
            v.validate_test(input_value, strict)
    else:
        assert v.isinstance_test(input_value, strict) is True
        assert v.validate_test(input_value, strict) == expected



#!/usr/bin/env python3
"""
This is used to generate flamegraphs, usage:

    perf record -g benchmarks/minimal.py

Then:

    perf script | stackcollapse-perf.pl | flamegraph.pl > flame.svg

Or just

    make flame

As per https://gist.github.com/KodrAus/97c92c07a90b1fdd6853654357fd557a
"""
import os
from pydantic_core import SchemaValidator, ValidationError
import json

size = 5
v = SchemaValidator(
    {
        'title': 'MyTestModel',
        'type': 'list',
        'items': {'type': 'model', 'fields': {f'f_{i}': {'type': 'str'} for i in range(size)}},
    }
)
# print(repr(v))

d = [{f'f_{i}': f'foobar_{i}' for i in range(size)} for _ in range(50)]
if os.getenv('JSON'):
    print('running validate_json...')
    j = json.dumps(d)
    for i in range(100_000):
        r = v.validate_json(j)
else:
    print('running validate_python...')
    for i in range(100_000):
        r = v.validate_python(d)
# debug(r)

try:
    r = v.validate_python({'name': 'John', 'age': 16, 'friends': [-1, 2, 3, -1], 'settings': {'a': 1.0, 'b': 2.0}})
except ValidationError as e:
    # print(e)
    pass



import re

import pytest

from pydantic_core import SchemaValidator, ValidationError

from ..conftest import Err, PyAndJson, plain_repr


@pytest.mark.parametrize(
    'input_value,expected',
    [
        (False, False),
        (True, True),
        (0, False),
        (0.0, False),
        (1, True),
        (1.0, True),
        ('yes', True),
        ('no', False),
        ('true', True),
        ('false', False),
        (
            'cheese',
            Err(
                'Input should be a valid boolean, '
                "unable to interpret input [kind=bool_parsing, input_value='cheese', input_type=str]"
            ),
        ),
        (2, Err('Input should be a valid boolean, unable to interpret input [kind=bool_parsing, input_value=2')),
        ([], Err('Input should be a valid boolean [kind=bool_type, input_value=[], input_type=list]')),
        (1.1, Err('Input should be a valid boolean [kind=bool_type, input_value=1.1, input_type=float]')),
        (2, Err('unable to interpret input [kind=bool_parsing, input_value=2, input_type=int]')),
        (2.0, Err('unable to interpret input [kind=bool_parsing, input_value=2.0, input_type=float]')),
    ],
)
def test_bool(py_and_json: PyAndJson, input_value, expected):
    v = py_and_json({'type': 'bool'})
    if isinstance(expected, Err):
        with pytest.raises(ValidationError, match=re.escape(expected.message)):
            v.validate_test(input_value)
        assert v.isinstance_test(input_value) is False
    else:
        assert v.validate_test(input_value) == expected
        assert v.isinstance_test(input_value) is True


def test_bool_strict(py_and_json: PyAndJson):
    v = py_and_json({'type': 'bool', 'strict': True})
    assert v.validate_test(True) is True
    error_message = "Input should be a valid boolean [kind=bool_type, input_value='true', input_type=str]"
    with pytest.raises(ValidationError, match=re.escape(error_message)):
        v.validate_test('true')


def test_bool_error():
    v = SchemaValidator({'type': 'bool'})

    with pytest.raises(ValidationError) as exc_info:
        v.validate_python('wrong')

    assert str(exc_info.value) == (
        '1 validation error for bool\n'
        '  Input should be a valid boolean, '
        "unable to interpret input [kind=bool_parsing, input_value='wrong', input_type=str]"
    )
    assert exc_info.value.errors() == [
        {
            'kind': 'bool_parsing',
            'loc': [],
            'message': 'Input should be a valid boolean, unable to interpret input',
            'input_value': 'wrong',
        }
    ]


def test_bool_repr():
    v = SchemaValidator({'type': 'bool'})
    assert plain_repr(v) == 'SchemaValidator(name="bool",validator=Bool(BoolValidator{strict:false}))'
    v = SchemaValidator({'type': 'bool', 'strict': True})
    assert plain_repr(v) == 'SchemaValidator(name="bool",validator=Bool(BoolValidator{strict:true}))'


def test_bool_key(py_and_json: PyAndJson):
    v = py_and_json({'type': 'dict', 'keys_schema': 'bool', 'values_schema': 'int'})
    assert v.validate_test({True: 1, False: 2}) == {True: 1, False: 2}
    assert v.validate_test({'true': 1, 'off': 2}) == {True: 1, False: 2}
    assert v.validate_test({'true': 1, 'off': 2}, strict=False) == {True: 1, False: 2}
    with pytest.raises(ValidationError, match='Input should be a valid boolean'):
        v.validate_test({'true': 1, 'off': 2}, strict=True)



import pytest

from pydantic_core import SchemaValidator, ValidationError


def func():
    return 42


class Foo:
    pass


class CallableClass:
    def __call__(self, *args, **kwargs):
        pass


def test_callable():
    v = SchemaValidator({'type': 'callable'})
    assert v.validate_python(func) == func
    assert v.isinstance_python(func) is True
    assert v.isinstance_json('"func"') is False

    with pytest.raises(ValidationError) as exc_info:
        v.validate_python(42)

    assert exc_info.value.errors() == [
        {'kind': 'callable_type', 'loc': [], 'message': 'Input should be callable', 'input_value': 42}
    ]


@pytest.mark.parametrize(
    'input_value,expected',
    [
        (func, True),
        (lambda: 42, True),
        (lambda x: 2 * 42, True),
        (dict, True),
        (Foo, True),
        (Foo(), False),
        (4, False),
        ('ddd', False),
        ([], False),
        ((1,), False),
        (CallableClass, True),
        (CallableClass(), True),
    ],
)
def test_callable_cases(input_value, expected):
    v = SchemaValidator({'type': 'callable'})
    assert v.isinstance_python(input_value) == expected


def test_repr():
    v = SchemaValidator({'type': 'union', 'choices': ['int', 'callable']})
    assert v.isinstance_python(4) is True
    assert v.isinstance_python(func) is True
    assert v.isinstance_python('foo') is False

    with pytest.raises(ValidationError, match=r'callable\s+Input should be callable'):
        v.validate_python('foo')



import pytest

from pydantic_core import SchemaValidator, ValidationError


@pytest.mark.parametrize(
    'input_value,output_value',
    [('false', False), ('true', True), ('0', False), ('1', True), ('"yes"', True), ('"no"', False)],
)
def test_bool(input_value, output_value):
    v = SchemaValidator({'type': 'bool'})
    assert v.validate_json(input_value) == output_value


@pytest.mark.parametrize('input_value', ['[1, 2, 3]', b'[1, 2, 3]', bytearray(b'[1, 2, 3]')])
def test_input_types(input_value):
    v = SchemaValidator({'type': 'list', 'items_schema': {'type': 'int'}})
    assert v.validate_json(input_value) == [1, 2, 3]


def test_input_type_invalid():
    v = SchemaValidator({'type': 'list', 'items_schema': {'type': 'int'}})
    with pytest.raises(TypeError, match='^JSON input should be str, bytes or bytearray, not list$'):
        v.validate_json([])


def test_null():
    assert SchemaValidator({'type': 'none'}).validate_json('null') is None


def test_str():
    s = SchemaValidator({'type': 'str'})
    assert s.validate_json('"foobar"') == 'foobar'
    with pytest.raises(ValidationError, match=r'Input should be a valid string \[kind=str_type,'):
        s.validate_json('false')
    with pytest.raises(ValidationError, match=r'Input should be a valid string \[kind=str_type,'):
        s.validate_json('123')


@pytest.mark.parametrize(
    'input_value,output_value',
    [('123.4', 123.4), ('123.0', 123.0), ('123', 123.0), ('"123.4"', 123.4), ('"123.0"', 123.0), ('"123"', 123.0)],
)
def test_float(input_value, output_value):
    v = SchemaValidator({'type': 'float'})
    assert v.validate_json(input_value) == output_value


def test_model():
    v = SchemaValidator(
        {
            'type': 'typed-dict',
            'fields': {'field_a': {'schema': {'type': 'str'}}, 'field_b': {'schema': {'type': 'int'}}},
        }
    )

    # language=json
    input_str = '{"field_a": "abc", "field_b": 1}'
    assert v.validate_json(input_str) == {'field_a': 'abc', 'field_b': 1}


def test_float_no_remainder():
    v = SchemaValidator({'type': 'int'})
    assert v.validate_json('123.0') == 123


def test_error_loc():
    v = SchemaValidator(
        {
            'type': 'typed-dict',
            'return_fields_set': True,
            'fields': {'field_a': {'schema': {'type': 'list', 'items_schema': {'type': 'int'}}}},
            'extra_validator': {'type': 'int'},
            'extra_behavior': 'allow',
        }
    )

    # assert v.validate_json('{"field_a": [1, 2, "3"]}') == ({'field_a': [1, 2, 3]}, {'field_a'})

    with pytest.raises(ValidationError) as exc_info:
        v.validate_json('{"field_a": [1, 2, "wrong"]}')
    assert exc_info.value.errors() == [
        {
            'kind': 'int_parsing',
            'loc': ['field_a', 2],
            'message': 'Input should be a valid integer, unable to parse string as an integer',
            'input_value': 'wrong',
        }
    ]


def test_dict():
    v = SchemaValidator({'type': 'dict', 'keys_schema': {'type': 'int'}, 'values_schema': {'type': 'int'}})
    assert v.validate_json('{"1": 2, "3": 4}') == {1: 2, 3: 4}


def test_dict_any_value():
    v = SchemaValidator({'type': 'dict', 'keys_schema': {'type': 'str'}})
    assert v.validate_json('{"1": 1, "2": "a", "3": null}') == {'1': 1, '2': 'a', '3': None}


def test_invalid_json():
    v = SchemaValidator({'type': 'bool'})

    with pytest.raises(ValidationError) as exc_info:
        v.validate_json('"foobar')
    assert exc_info.value.errors() == [
        {
            'kind': 'invalid_json',
            'loc': [],
            'message': 'Invalid JSON: EOF while parsing a string at line 1 column 7',
            'input_value': '"foobar',
            'context': {'error': 'EOF while parsing a string at line 1 column 7'},
        }
    ]
    with pytest.raises(ValidationError) as exc_info:
        v.validate_json('[1,\n2,\n3,]')
    assert exc_info.value.errors() == [
        {
            'kind': 'invalid_json',
            'loc': [],
            'message': 'Invalid JSON: trailing comma at line 3 column 3',
            'input_value': '[1,\n2,\n3,]',
            'context': {'error': 'trailing comma at line 3 column 3'},
        }
    ]



import json
import platform
import re
from datetime import date, datetime, time, timedelta, timezone, tzinfo
from decimal import Decimal

import pytest
import pytz

from pydantic_core import SchemaError, SchemaValidator, ValidationError

from ..conftest import Err, PyAndJson


@pytest.mark.parametrize(
    'input_value,expected',
    [
        (datetime(2022, 6, 8, 12, 13, 14), datetime(2022, 6, 8, 12, 13, 14)),
        (date(2022, 6, 8), datetime(2022, 6, 8)),
        ('2022-06-08T12:13:14', datetime(2022, 6, 8, 12, 13, 14)),
        (b'2022-06-08T12:13:14', datetime(2022, 6, 8, 12, 13, 14)),
        (b'2022-06-08T12:13:14Z', datetime(2022, 6, 8, 12, 13, 14, tzinfo=timezone.utc)),
        ((1,), Err('Input should be a valid datetime [kind=datetime_type')),
        (time(1, 2, 3), Err('Input should be a valid datetime [kind=datetime_type')),
        (Decimal('1654646400'), datetime(2022, 6, 8)),
        (Decimal('1654646400.123456'), datetime(2022, 6, 8, 0, 0, 0, 123456)),
        (Decimal('1654646400.1234564'), datetime(2022, 6, 8, 0, 0, 0, 123456)),
        (Decimal('1654646400.1234568'), datetime(2022, 6, 8, 0, 0, 0, 123457)),
        (253_402_300_800_000, Err('should be a valid datetime, dates after 9999 are not supported as unix timestamps')),
        (-20_000_000_000, Err('should be a valid datetime, dates before 1600 are not supported as unix timestamps')),
    ],
)
def test_datetime(input_value, expected):
    v = SchemaValidator({'type': 'datetime'})
    if isinstance(expected, Err):
        with pytest.raises(ValidationError, match=re.escape(expected.message)):
            v.validate_python(input_value)
    else:
        output = v.validate_python(input_value)
        assert output == expected


@pytest.mark.parametrize(
    'input_value,expected',
    [
        (datetime(2022, 6, 8, 12, 13, 14), datetime(2022, 6, 8, 12, 13, 14)),
        (date(2022, 6, 8), Err('Input should be a valid datetime [kind=datetime_type')),
        ('2022-06-08T12:13:14', Err('Input should be a valid datetime [kind=datetime_type')),
        (b'2022-06-08T12:13:14', Err('Input should be a valid datetime [kind=datetime_type')),
        (time(1, 2, 3), Err('Input should be a valid datetime [kind=datetime_type')),
        (1654646400, Err('Input should be a valid datetime [kind=datetime_type')),
        (Decimal('1654646400'), Err('Input should be a valid datetime [kind=datetime_type')),
    ],
)
def test_datetime_strict(input_value, expected):
    v = SchemaValidator({'type': 'datetime', 'strict': True})
    if isinstance(expected, Err):
        with pytest.raises(ValidationError, match=re.escape(expected.message)):
            v.validate_python(input_value)
    else:
        output = v.validate_python(input_value)
        assert output == expected


def test_keep_tz():
    tz = pytz.timezone('Europe/London')
    dt = tz.localize(datetime(2022, 6, 14, 12, 13, 14))
    v = SchemaValidator({'type': 'datetime'})

    output = v.validate_python(dt)
    assert output == dt

    # dst object is unaffected by validation
    assert output.tzinfo.dst(datetime(2022, 6, 1)) == timedelta(seconds=3600)
    assert output.tzinfo.dst(datetime(2022, 1, 1)) == timedelta(seconds=0)


def test_keep_tz_bound():
    tz = pytz.timezone('Europe/London')
    dt = tz.localize(datetime(2022, 6, 14, 12, 13, 14))
    v = SchemaValidator({'type': 'datetime', 'gt': datetime(2022, 1, 1)})

    output = v.validate_python(dt)
    assert output == dt

    # dst object is unaffected by validation
    assert output.tzinfo.dst(datetime(2022, 6, 1)) == timedelta(hours=1)
    assert output.tzinfo.dst(datetime(2022, 1, 1)) == timedelta(0)

    with pytest.raises(ValidationError, match=r'Input should be greater than 2022-01-01T00:00:00 \[kind=greater_than'):
        v.validate_python(tz.localize(datetime(2021, 6, 14)))


@pytest.mark.parametrize(
    'input_value,expected',
    [
        ('2022-06-08T12:13:14', datetime(2022, 6, 8, 12, 13, 14)),
        ('2022-06-08T12:13:14Z', datetime(2022, 6, 8, 12, 13, 14, tzinfo=timezone.utc)),
        (
            '2022-06-08T12:13:14+12:15',
            datetime(2022, 6, 8, 12, 13, 14, tzinfo=timezone(timedelta(hours=12, minutes=15))),
        ),
        (
            '2022-06-08T12:13:14+23:59',
            datetime(2022, 6, 8, 12, 13, 14, tzinfo=timezone(timedelta(hours=23, minutes=59))),
        ),
        (1655205632, datetime(2022, 6, 14, 11, 20, 32)),
        (1655205632.331557, datetime(2022, 6, 14, 11, 20, 32, microsecond=331557)),
        (
            '2022-06-08T12:13:14+24:00',
            Err('Input should be a valid datetime, timezone offset must be less than 24 hours [kind=datetime_parsing,'),
        ),
        (True, Err('Input should be a valid datetime [kind=datetime_type')),
        (None, Err('Input should be a valid datetime [kind=datetime_type')),
        ([1, 2, 3], Err('Input should be a valid datetime [kind=datetime_type')),
    ],
)
def test_datetime_json(py_and_json: PyAndJson, input_value, expected):
    v = py_and_json({'type': 'datetime'})
    if isinstance(expected, Err):
        with pytest.raises(ValidationError, match=re.escape(expected.message)):
            v.validate_test(input_value)
    else:
        output = v.validate_test(input_value)
        assert output == expected


@pytest.mark.parametrize(
    'input_value,expected',
    [
        ('2022-06-08T12:13:14', datetime(2022, 6, 8, 12, 13, 14)),
        ('2022-06-08T12:13:14Z', datetime(2022, 6, 8, 12, 13, 14, tzinfo=timezone.utc)),
        (123, Err('Input should be a valid datetime [kind=datetime_type')),
        (123.4, Err('Input should be a valid datetime [kind=datetime_type')),
        (True, Err('Input should be a valid datetime [kind=datetime_type')),
    ],
)
def test_datetime_strict_json(input_value, expected):
    v = SchemaValidator({'type': 'datetime', 'strict': True})
    if isinstance(expected, Err):
        with pytest.raises(ValidationError, match=re.escape(expected.message)):
            v.validate_json(json.dumps(input_value))
    else:
        output = v.validate_json(json.dumps(input_value))
        assert output == expected


def test_custom_timezone_repr():
    output = SchemaValidator({'type': 'datetime'}).validate_python('2022-06-08T12:13:14-12:15')
    assert output == datetime(2022, 6, 8, 12, 13, 14, tzinfo=timezone(timedelta(hours=-12, minutes=-15)))
    assert output.tzinfo.utcoffset(output) == timedelta(hours=-12, minutes=-15)
    assert output.tzinfo.dst(output) is None
    assert output.tzinfo.tzname(output) == '-12:15'
    assert str(output.tzinfo) == '-12:15'
    assert repr(output.tzinfo) == 'TzInfo(-12:15)'


def test_custom_timezone_utc_repr():
    output = SchemaValidator({'type': 'datetime'}).validate_python('2022-06-08T12:13:14Z')
    assert output == datetime(2022, 6, 8, 12, 13, 14, tzinfo=timezone(timedelta(0)))
    assert output.tzinfo.utcoffset(output) == timedelta(0)
    assert output.tzinfo.dst(output) is None
    assert output.tzinfo.tzname(output) == 'UTC'
    assert str(output.tzinfo) == 'UTC'
    assert repr(output.tzinfo) == 'TzInfo(UTC)'


def test_tz_comparison():
    tz = pytz.timezone('Europe/London')
    uk_3pm = tz.localize(datetime(2022, 1, 1, 15, 0, 0))

    # two times are the same instant, therefore le and ge are both ok
    v = SchemaValidator({'type': 'datetime', 'le': uk_3pm}).validate_python('2022-01-01T16:00:00+01:00')
    assert v == datetime(2022, 1, 1, 16, 0, 0, tzinfo=timezone(timedelta(hours=1)))

    v = SchemaValidator({'type': 'datetime', 'ge': uk_3pm}).validate_python('2022-01-01T16:00:00+01:00')
    assert v == datetime(2022, 1, 1, 16, 0, 0, tzinfo=timezone(timedelta(hours=1)))

    # but not gt
    with pytest.raises(ValidationError, match=r'Input should be greater than 2022-01-01T15:00:00Z \[kind=greater_than'):
        SchemaValidator({'type': 'datetime', 'gt': uk_3pm}).validate_python('2022-01-01T16:00:00+01:00')


def test_custom_tz():
    class CustomTz(tzinfo):
        def utcoffset(self, _dt):
            return None

        def dst(self, _dt):
            return None

        def tzname(self, _dt):
            return 'CustomTZ'

    schema = SchemaValidator({'type': 'datetime', 'gt': datetime(2022, 1, 1, 15, 0, 0)})

    dt = datetime(2022, 1, 1, 16, 0, 0, tzinfo=CustomTz())
    outcome = schema.validate_python(dt)
    assert outcome == dt


def test_custom_invalid_tz():
    class CustomTz(tzinfo):
        # utcoffset is not implemented!

        def tzname(self, _dt):
            return 'CustomTZ'

    schema = SchemaValidator({'type': 'datetime', 'gt': datetime(2022, 1, 1, 15, 0, 0)})

    dt = datetime(2022, 1, 1, 16, 0, 0, tzinfo=CustomTz())
    # perhaps this should be a ValidationError? but we don't catch other errors
    with pytest.raises(ValidationError) as excinfo:
        schema.validate_python(dt)

    # exception messages differ between python and pypy
    if platform.python_implementation() == 'PyPy':
        error_message = 'NotImplementedError: tzinfo subclass must override utcoffset()'
    else:
        error_message = 'NotImplementedError: a tzinfo subclass must implement utcoffset()'

    assert excinfo.value.errors() == [
        {
            'kind': 'datetime_object_invalid',
            'loc': [],
            'message': f'Invalid datetime object, got {error_message}',
            'input_value': dt,
            'context': {'error': error_message},
        }
    ]


def test_dict_py():
    v = SchemaValidator({'type': 'dict', 'keys_schema': 'datetime', 'values_schema': 'int'})
    assert v.validate_python({datetime(2000, 1, 1): 2, datetime(2000, 1, 2): 4}) == {
        datetime(2000, 1, 1): 2,
        datetime(2000, 1, 2): 4,
    }


def test_dict(py_and_json: PyAndJson):
    v = py_and_json({'type': 'dict', 'keys_schema': 'datetime', 'values_schema': 'int'})
    assert v.validate_test({'2000-01-01T00:00': 2, '2000-01-02T00:00': 4}) == {
        datetime(2000, 1, 1): 2,
        datetime(2000, 1, 2): 4,
    }


def test_union():
    v = SchemaValidator({'type': 'union', 'choices': ['str', 'datetime']})
    assert v.validate_python('2022-01-02T00:00') == '2022-01-02T00:00'
    assert v.validate_python(datetime(2022, 1, 2)) == datetime(2022, 1, 2)

    v = SchemaValidator({'type': 'union', 'choices': ['datetime', 'str']})
    assert v.validate_python('2022-01-02T00:00') == '2022-01-02T00:00'
    assert v.validate_python(datetime(2022, 1, 2)) == datetime(2022, 1, 2)


def test_invalid_constraint():
    with pytest.raises(SchemaError, match='datetime -> gt\n  Input should be a valid datetime'):
        SchemaValidator({'type': 'datetime', 'gt': 'foobar'})



import re
from decimal import Decimal
from typing import Any, Dict

import pytest

from pydantic_core import SchemaValidator, ValidationError

from ..conftest import Err, PyAndJson, plain_repr


@pytest.mark.parametrize(
    'input_value,expected',
    [
        (False, 0),
        (True, 1),
        (0, 0),
        ('0', 0),
        (1, 1),
        (42, 42),
        ('42', 42),
        (42.0, 42),
        (int(1e10), int(1e10)),
        pytest.param(
            12.5,
            Err('Input should be a valid integer, got a number with a fractional part [kind=int_from_float'),
            id='float-remainder',
        ),
        pytest.param(
            'wrong',
            Err('Input should be a valid integer, unable to parse string as an integer [kind=int_parsing'),
            id='string',
        ),
        pytest.param(None, Err('Input should be a valid integer [kind=int_type'), id='list'),
        pytest.param([1, 2], Err('Input should be a valid integer [kind=int_type'), id='list'),
    ],
)
def test_int_py_and_json(py_and_json: PyAndJson, input_value, expected):
    v = py_and_json({'type': 'int'})
    if isinstance(expected, Err):
        with pytest.raises(ValidationError, match=re.escape(expected.message)):
            v.validate_test(input_value)
    else:
        output = v.validate_test(input_value)
        assert output == expected
        assert isinstance(output, int)


@pytest.mark.parametrize(
    'input_value,expected',
    [
        (Decimal('1'), 1),
        (Decimal('1.0'), 1),
        pytest.param(
            Decimal('1.001'),
            Err(
                'Input should be a valid integer, got a number with a fractional part '
                "[kind=int_from_float, input_value=Decimal('1.001'), input_type=Decimal]"
            ),
            id='decimal-remainder',
        ),
        pytest.param(
            (1, 2),
            Err('Input should be a valid integer [kind=int_type, input_value=(1, 2), input_type=tuple]'),
            id='tuple',
        ),
    ],
)
def test_int(input_value, expected):
    v = SchemaValidator({'type': 'int'})
    if isinstance(expected, Err):
        with pytest.raises(ValidationError, match=re.escape(expected.message)):
            v.validate_python(input_value)
    else:
        output = v.validate_python(input_value)
        assert output == expected
        assert isinstance(output, int)


@pytest.mark.parametrize(
    'input_value,expected',
    [
        (0, 0),
        (1, 1),
        (42, 42),
        pytest.param(
            42.0,
            Err('Input should be a valid integer [kind=int_type, input_value=42.0, input_type=float]'),
            id='float-exact',
        ),
        pytest.param(
            42.5,
            Err('Input should be a valid integer [kind=int_type, input_value=42.5, input_type=float]'),
            id='float-remainder',
        ),
        pytest.param(
            '42', Err("Input should be a valid integer [kind=int_type, input_value='42', input_type=str]"), id='string'
        ),
        pytest.param(
            True, Err('Input should be a valid integer [kind=int_type, input_value=True, input_type=bool]'), id='bool'
        ),
    ],
)
def test_int_strict(py_and_json: PyAndJson, input_value, expected):
    v = py_and_json({'type': 'int', 'strict': True})
    if isinstance(expected, Err):
        with pytest.raises(ValidationError, match=re.escape(expected.message)):
            v.validate_test(input_value)
    else:
        assert v.validate_test(input_value) == expected


@pytest.mark.parametrize(
    'kwargs,input_value,expected',
    [
        ({}, 0, 0),
        ({}, '123.000', 123),
        ({'ge': 0}, 0, 0),
        (
            {'ge': 0},
            -1,
            Err(
                'Input should be greater than or equal to 0 '
                '[kind=greater_than_equal, input_value=-1, input_type=int]'
            ),
        ),
        ({'gt': 0}, 1, 1),
        ({'gt': 0}, 0, Err('Input should be greater than 0 [kind=greater_than, input_value=0, input_type=int]')),
        ({'le': 0}, 0, 0),
        ({'le': 0}, -1, -1),
        ({'le': 0}, 1, Err('Input should be less than or equal to 0')),
        ({'lt': 0}, 0, Err('Input should be less than 0')),
        ({'lt': 0}, 1, Err('Input should be less than 0')),
        ({'multiple_of': 5}, 15, 15),
        ({'multiple_of': 5}, 6, Err('Input should be a multiple of 5')),
    ],
    ids=repr,
)
def test_int_kwargs(py_and_json: PyAndJson, kwargs: Dict[str, Any], input_value, expected):
    v = py_and_json({'type': 'int', **kwargs})
    if isinstance(expected, Err):
        with pytest.raises(ValidationError, match=re.escape(expected.message)) as exc_info:
            v.validate_test(input_value)

        errors = exc_info.value.errors()
        assert len(errors) == 1
        if 'context' in errors[0]:
            assert errors[0]['context'] == kwargs
    else:
        output = v.validate_test(input_value)
        assert output == expected
        assert isinstance(output, int)


def test_union_int(py_and_json: PyAndJson):
    v = py_and_json({'type': 'union', 'choices': [{'type': 'int', 'strict': True}, {'type': 'int', 'multiple_of': 7}]})
    assert v.validate_test('14') == 14
    assert v.validate_test(5) == 5
    with pytest.raises(ValidationError) as exc_info:
        v.validate_test('5')

    assert exc_info.value.errors() == [
        {'kind': 'int_type', 'loc': ['int'], 'message': 'Input should be a valid integer', 'input_value': '5'},
        {
            'kind': 'multiple_of',
            'loc': ['constrained-int'],
            'message': 'Input should be a multiple of 7',
            'input_value': '5',
            'context': {'multiple_of': 7},
        },
    ]


def test_union_int_simple(py_and_json: PyAndJson):
    v = py_and_json({'type': 'union', 'choices': [{'type': 'int'}]})
    assert v.validate_test('5') == 5
    with pytest.raises(ValidationError) as exc_info:
        v.validate_test('xxx')

    assert exc_info.value.errors() == [
        {
            'kind': 'int_parsing',
            'loc': ['int'],
            'message': 'Input should be a valid integer, unable to parse string as an integer',
            'input_value': 'xxx',
        }
    ]


def test_int_repr():
    v = SchemaValidator({'type': 'int'})
    assert plain_repr(v) == 'SchemaValidator(name="int",validator=Int(IntValidator{strict:false}))'
    v = SchemaValidator({'type': 'int', 'strict': True})
    assert plain_repr(v) == 'SchemaValidator(name="int",validator=Int(IntValidator{strict:true}))'
    v = SchemaValidator({'type': 'int', 'multiple_of': 7})
    assert plain_repr(v).startswith('SchemaValidator(name="constrained-int",validator=ConstrainedInt(')


def test_long_int(py_and_json: PyAndJson):
    v = py_and_json({'type': 'int'})

    with pytest.raises(ValidationError) as exc_info:
        v.validate_test('1' * 400)

    assert exc_info.value.errors() == [
        {
            'kind': 'int_nan',
            'loc': [],
            'message': 'Input should be a valid integer, got infinity',
            'input_value': '1' * 400,
            'context': {'nan_value': 'infinity'},
        }
    ]
    assert repr(exc_info.value) == (
        '1 validation error for int\n'
        '  Input should be a valid integer, got infinity '
        '[kind=int_nan, '
        "input_value='111111111111111111111111...11111111111111111111111', input_type=str]"
    )


def test_int_nan(py_and_json: PyAndJson):
    v = py_and_json({'type': 'int'})

    with pytest.raises(ValidationError, match='Input should be a valid integer, got negative infinity'):
        v.validate_test('-' + '1' * 400)

    with pytest.raises(ValidationError, match='Input should be a valid integer, got NaN'):
        v.validate_test('nan')


def test_int_key(py_and_json: PyAndJson):
    v = py_and_json({'type': 'dict', 'keys_schema': 'int', 'values_schema': 'int'})
    assert v.validate_test({'1': 1, '2': 2}) == {1: 1, 2: 2}
    with pytest.raises(ValidationError, match='Input should be a valid integer'):
        v.validate_test({'1': 1, '2': 2}, strict=True)



"""
General benchmarks that attempt to cover all field types, through by no means all uses of all field types.
"""
import json
import sys
from datetime import date, datetime, time

import pytest

from pydantic_core import SchemaValidator, ValidationError

from .complete_schema import input_data_lax, input_data_strict, input_data_wrong, pydantic_model, schema
from .test_micro_benchmarks import skip_pydantic

pytestmark = pytest.mark.skipif(sys.version_info < (3, 10), reason='requires python3.10 or higher')


def test_complete_valid():
    lax_schema = schema()
    class_type = lax_schema['class_type']
    lax_validator = SchemaValidator(lax_schema)
    output = lax_validator.validate_python(input_data_lax())
    assert isinstance(output, class_type)
    assert len(output.__fields_set__) == 39
    output_dict = output.__dict__
    assert output_dict == {
        'field_str': 'fo',
        'field_str_con': 'fooba',
        'field_int': 1,
        'field_int_con': 8,
        'field_float': 1.0,
        'field_float_con': 10.0,
        'field_bool': True,
        'field_bytes': b'foobar',
        'field_bytes_con': b'foobar',
        'field_date': date(2010, 2, 3),
        'field_date_con': date(2020, 1, 1),
        'field_time': time(12, 0),
        'field_time_con': time(12, 0),
        'field_datetime': datetime(2020, 1, 1, 12, 13, 14),
        'field_datetime_con': datetime(2020, 1, 1),
        'field_list_any': ['a', b'b', True, 1.0, None] * 10,
        'field_list_str': ['a', 'b', 'c'] * 10,
        'field_list_str_con': ['a', 'b', 'c'] * 10,
        'field_set_any': {'a', b'b', True, 1.0, None},
        'field_set_int': set(range(100)),
        'field_set_int_con': set(range(42)),
        'field_frozenset_any': frozenset({'a', b'b', True, 1.0, None}),
        'field_frozenset_bytes': frozenset([f'{i}'.encode() for i in range(100)]),
        'field_frozenset_bytes_con': frozenset([f'{i}'.encode() for i in range(42)]),
        'field_tuple_var_len_any': ('a', b'b', True, 1.0, None),
        'field_tuple_var_len_float': tuple((i + 0.5 for i in range(100))),
        'field_tuple_var_len_float_con': tuple((i + 0.5 for i in range(42))),
        'field_tuple_fix_len': ('a', 1, 1.0, True),
        'field_dict_any': {'a': 'b', 1: True, 1.0: 1.0},
        'field_dict_str_float': {f'{i}': i + 0.5 for i in range(100)},
        'field_literal_1_int': 1,
        'field_literal_1_str': 'foobar',
        'field_literal_mult_int': 3,
        'field_literal_mult_str': 'foo',
        'field_literal_assorted': 'foo',
        'field_list_nullable_int': [1, None, 2, None, 3, None, 4, None],
        'field_union': {'field_str': 'foo', 'field_int': 1, 'field_float': 1.0},
        'field_functions_model': {
            'field_before': 'foo Changed',
            'field_after': 'foo Changed',
            'field_wrap': 'Input foo Changed',
            'field_plain': 'foo Changed',
        },
        'field_recursive': {
            'name': 'foo',
            'sub_branch': {'name': 'bar', 'sub_branch': {'name': 'baz', 'sub_branch': None}},
        },
    }

    strict_validator = SchemaValidator(schema(strict=True))
    output2 = strict_validator.validate_python(input_data_strict())
    assert output_dict == output2.__dict__

    model = pydantic_model()
    if model is None:
        print('pydantic is not installed, skipping pydantic tests')
        return

    output_pydantic = model.parse_obj(input_data_lax())
    assert output_pydantic.dict() == output_dict


def test_complete_invalid():
    lax_schema = schema()
    lax_validator = SchemaValidator(lax_schema)
    with pytest.raises(ValidationError) as exc_info:
        lax_validator.validate_python(input_data_wrong())
    assert len(exc_info.value.errors()) == 738

    model = pydantic_model()
    if model is None:
        print('pydantic is not installed, skipping pydantic tests')
        return

    from pydantic import ValidationError as PydanticValidationError

    with pytest.raises(PydanticValidationError) as exc_info:
        model.parse_obj(input_data_wrong())
    assert len(exc_info.value.errors()) == 530


@pytest.mark.benchmark(group='complete')
def test_complete_core_lax(benchmark):
    v = SchemaValidator(schema())
    benchmark(v.validate_python, input_data_lax())


@pytest.mark.benchmark(group='complete')
def test_complete_core_strict(benchmark):
    v = SchemaValidator(schema(strict=True))
    benchmark(v.validate_python, input_data_strict())


@skip_pydantic
@pytest.mark.benchmark(group='complete')
def test_complete_pyd(benchmark):
    model = pydantic_model()
    assert model is not None
    benchmark(model.parse_obj, input_data_lax())


@pytest.mark.benchmark(group='complete-wrong')
def test_complete_core_error(benchmark):
    v = SchemaValidator(schema())
    data = input_data_wrong()

    @benchmark
    def f():
        try:
            v.validate_python(data)
        except ValueError:
            pass
        else:
            raise RuntimeError('expected ValueError')


@pytest.mark.benchmark(group='complete-wrong')
def test_complete_core_isinstance(benchmark):
    v = SchemaValidator(schema())
    data = input_data_wrong()
    assert v.isinstance_python(data) is False

    @benchmark
    def f():
        v.isinstance_python(data)


@skip_pydantic
@pytest.mark.benchmark(group='complete-wrong')
def test_complete_pyd_error(benchmark):
    model = pydantic_model()
    assert model is not None
    data = input_data_wrong()

    @benchmark
    def f():
        try:
            model.parse_obj(data)
        except ValueError:
            pass
        else:
            raise RuntimeError('expected ValueError')


def default_json_encoder(obj):
    if isinstance(obj, bytes):
        return obj.decode('utf-8')
    if isinstance(obj, (set, frozenset)):
        return list(obj)
    else:
        raise TypeError(f'Object of type {type(obj)} is not JSON serializable')


@pytest.mark.benchmark(group='complete-json')
def test_complete_core_json(benchmark):
    v = SchemaValidator(schema())
    json_data = json.dumps(input_data_lax(), default=default_json_encoder)
    benchmark(v.validate_json, json_data)


@skip_pydantic
@pytest.mark.benchmark(group='complete-json')
def test_complete_pyd_json(benchmark):
    model = pydantic_model()
    assert model is not None
    json_data = json.dumps(input_data_lax(), default=default_json_encoder)

    @benchmark
    def t():
        model.parse_raw(json_data, content_type='application/json')


@pytest.mark.benchmark(group='build')
def test_build_schema(benchmark):
    lax_schema = schema()
    benchmark(SchemaValidator, lax_schema)
